# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read AMSU-A and MHS MIRS NetCDF data files.

This reader is desgined for importing the Advanced Microwave Sounding Unit
(AMSU)-A and Microwave Humidity Sounder (HMS) files made available from NOAA MIRS.
These data files have a different file name convention
and and data structure from previous MSPPS.  Example of MIRIS files:

* NPR-MIRS-IMG_v11r4_ma1_s202101111916000_e202101112012000_c202101112047200.nc

    * (ma1 is for metop-B)
* NPR-MIRS-IMG_v11r4_ma2_s202101111715000_e202101111857000_c202101111941370.nc

    * (ma2 is for metop-A)
* NPR-MIRS-IMG_v11r4_n19_s202101111730000_e202101111916000_c202101112001590.nc

    * (NOAA-19)

AMSU-B was replaced with MHS in 2005, and now AMSU-A / MHS fly on newer NOAA birds (>17)
and METOP-A/B/C (METOP-A dead now).  AMSU-A channels are still channels 1-15, and MHS
channels (1-5) live in the same 16-20 slots that previously contained AMSU-B.

More information on the AMSU and MHS instruments may be found here
https://en.wikipedia.org/wiki/Advanced_microwave_sounding_unit
https://en.wikipedia.org/wiki/Microwave_humidity_sounder

AMSU-A channel information:

Chan  Frequency                 Polarization  #Bands  Sensitivity  Primary Function
Num   (GHz)                     (at nadir)            NEDT (K)
1     23.8                      vertical      1       0.30         Water Vapor Burden
2     31.4                      vertical      1       0.30         Water Vapor Burden
3     50.3                      vertical      1       0.40         Water Vapor Burden
4     52.8                      vertical      1       0.25         Water Vapor Burden
5     53.596 ± 0.115            horizontal    2       0.25         Tropospheric Temp
6     54.4                      horizontal    1       0.25         Tropospheric Temp
7     54.94                     vertical      1       0.25         Tropospheric Temp
8     55.5                      horizontal    1       0.25         Tropospheric Temp
9     57.290                    horizontal    1       0.25         Stratospheric Temp
10    57.290 ± 0.217            horizontal    2       0.40         Stratospheric Temp
11    57.290 ± 0.3222 ± 0.048   horizontal    4       0.40         Stratospheric Temp
12    57.290 ± 0.3222 ± 0.022   horizontal    4       0.60         Stratospheric Temp
13    57.290 ± 0.3222 ± 0.010   horizontal    4       0.80         Stratospheric Temp
14    57.290 ± 0.3222 ± 0.0045  horizontal    4       1.20         Stratospheric Temp
15    89.0                      vertical      1       0.50         Cloud Top/Snow

MHS channel information:

Chan  AMSU-B  Frequency       AMSU-B         Bandwidth  Sensitivity  Polarization
Num   Chan#  (GHz)            Freq(GHz)      (MHz)      NEDT (K)     (at nadir)
1     16      89.0            89.0 ± 0.9     2800       0.22         Vertical
2     17      157.0           150 ± 0.9      2800       0.34         Vertical
3     18      183.311 ± 1.00  183.31 ± 1.00  2 x 500    0.51         Horizontal
4     19      183.311 ± 3.00  183.31 ± 3.00  2 x 1000   0.40         Horizontal
5     20      190.311         183.31 ± 7.00  2200       0.46         Vertical

V1.0:  Initial version, NRL-MRY, January 26, 2021

Dataset information::

    Basic information on AMSU-A MHS MIRS product file
    index: 1     2    3    4      5      6    7
    freq: 23.8,31.4,50.3,52.799,53.595,54.4,54.941,
    55.499,57.29,57.29,57.29,57.29,57.29,57.29,
    index:   8     9      10    11   12   13     14
    freq: 55.499,57.29,57.29,57.29,57.29,57.29,57.29,
    index: 15  16  17    18      19     20
    freq: 89.,89.,157.,183.311,183.311,190.311

    dimensions:
          Scanline = 2370 ;
          Field_of_view = 90 ;
          P_Layer = 100 ;
          Channel = 20 ;
          Qc_dim = 4 ;
    variables:
          Freq(Channel): Central Frequencies (GHz)
          Polo(Channel): Polarizations
          ScanTime_year(Scanline): Calendar Year 20XX
          ScanTime_doy(Scanline: julian day 1-366
          ScanTime_month(Scanline): Calendar month 1-12
          ScanTime_dom(Scanline): Calendar day of the month 1-31
          ScanTime_hour(Scanline: hour of the day 0-23
          ScanTime_minute(Scanline): minute of the hour 0-59
          ScanTime_second(Scanline): second of the minute 0-59
          ScanTime_UTC(Scanline): Number of seconds since 00:00:00 UTC
          Orb_mode(Scanline): 0-ascending,1-descending
          Latitude(Scanline, Field_of_view):Latitude of the view (-90,90),
                                            unit: degree
          Longitude(Scanline, Field_of_view):Longitude of the view (-180,180),
                                             unit: degree
          Sfc_type(Scanline, Field_of_view):type of surface:0-ocean,1-sea ice,
                                            2-land,3-snow
          Atm_type(Scanline, Field_of_view): type of atmosphere:currently missing
                                          ( note: not needed for geoips products)
          Qc(Scanline, Field_of_view, Qc_dim): Qc: 0-good, 1-usable with problem,
                                                   2-bad
          ChiSqr(Scanline, Field_of_view): Convergence rate: <3-good,>10-bad
          LZ_angle(Scanline, Field_of_view): Local Zenith Angle degree
          RAzi_angle(Scanline, Field_of_view):Relative Azimuth Angle 0-360 degree
          SZ_angle(Scanline, Field_of_view):Solar Zenith Angle (-90,90) degree
          BT(Scanline, Field_of_view, Channel): Channel Temperature (K)
          YM(Scanline, Field_of_view, Channel): UnCorrected Channel Temperature(K)
          ChanSel(Scanline, Field_of_view, Channel):Channels Selection Used in
                                                    Retrieval
          TPW(Scanline, Field_of_view): Total Precipitable Water (mm)
          CLW(Scanline, Field_of_view):Cloud liquid Water (mm)
          RWP(Scanline, Field_of_view): Rain Water Path (mm)
          LWP(Scanline, Field_of_view): Liquid Water Path (mm)
          SWP(Scanline, Field_of_view): Snow Water Path (mm)
          IWP(Scanline, Field_of_view): Ice Water Path (mm)
          GWP(Scanline, Field_of_view): Graupel Water Path (mm)
          RR(Scanline, Field_of_view): Rain Rate (mm/hr)
          Snow(Scanline, Field_of_view): Snow Cover (range: 0-1) i.e., 1 -> 100%
          SWE(Scanline, Field_of_view): Snow Water Equivalent (cm)
          SnowGS(Scanline, Field_of_view):Snow Grain Size (mm)
          SIce(Scanline, Field_of_view):Sea Ice Concentration (%)
          SIce_MY(Scanline, Field_of_view): Multi-Year Sea Ice Concentration (%)
          SIce_FY(Scanline, Field_of_view): First-Year Sea Ice Concentration (%)
          TSkin(Scanline, Field_of_view): Skin Temperature (K)
          SurfP(Scanline, Field_of_view): Surface Pressure (mb)
          Emis(Scanline, Field_of_view, Channel):Channel Emissivity
                                          (unit:1,  Emis:scale_factor = 0.0001)
          SFR(Scanline, Field_of_view): Snow Fall Rate in mm/hr
          CldTop(Scanline, Field_of_view): Cloud Top Pressure (scale_factor = 0.1)
          CldBase(Scanline, Field_of_view): Cloud Base Pressure
                                            (scale_factor = 0.1)
          CldThick(Scanline, Field_of_view): Cloud Thickness (scale_factor = 0.1)
          PrecipType(Scanline, Field_of_view): Precipitation Type (Frozen/Liquid)
          RFlag(Scanline, Field_of_view): Rain Flag
          SurfM(Scanline, Field_of_view): Surface Moisture (scale_factor = 0.1)
          WindSp(Scanline, Field_of_view): Wind Speed (m/s) (scale_factor = 0.01
          WindDir(Scanline, Field_of_view: Wind Direction (scale_factor = 0.01)
          WindU(Scanline, Field_of_view):U-direction Wind Speed (m/s)
                                        (scale_factor = 0.01)
          WindV(Scanline, Field_of_view: V-direction Wind Speed (m/s)
                                        (scale_factor = 0.01)
          Prob_SF(Scanline, Field_of_view): Probability of falling snow (%)

Additional info::

    Variables (nscan, npix):  npix=90 pixels per scan; nscan: vary with orbit
    chan-1 AT:  89 GHz              as ch16    anttenna temperature at V-pol
                                               FOV 16km at nadir
    chan-2 AT:  150 (157) GHz       as ch17 the number in bracket is for MHS
                                           from metops at V-pol, 16km at nadir
    chan-3 AT:  183.31 +-1 GHz          as ch18    at H-pol, 16km
    chan-4 AT:  183.31 +-3 GHz          as ch19    at H-pol, 16km
    chan-5 AT:  183.31 +-7 (190.3) GHz  as ch20    at V-pol, 16km
    lat:     -90, 90     deg
    lon:     -180, 180   deg
    RR:  surface rainrate (mm/hr)
    Snow: surafce snow cover
    IWP:  ice water path
    SWE:  snow water equvelent
    Sfc_type:  surface type
    Orbit_mode:   -1: ascending, 1: decending, 2: both
    SFR:  snowfall rate (unit?)
    LZ_angle:  local zinath angle (deg)
    SZ_angle:  solar zinath angle (deg)
"""
# Python Standard Libraries
from datetime import datetime
import logging
import os

# Third-Party Libraries
from collections import defaultdict
import h5py
import numpy as np
import pandas as pd
import xarray as xr

LOG = logging.getLogger(__name__)

# Selected variables for TC products (add more if needed)
VARLIST = [
    "BT",
    "Latitude",
    "Longitude",
    "Freq",
    "RR",
    "TPW",
    "CLW",
    "RWP",
    "LWP",
    "SWP",
    "IWP",
    "GWP",
    "Snow",
    "SWE",
    "SIce",
    "SIce_MY",
    "SIce_FY",
    "TSkin",
    "SurfP",
    "Emis",
    "Sfc_type",
    "SFR",
    "CldTop",
    "CldBase",
    "CldThick",
    "WindSp",
    "WindDir",
    "WindU",
    "WindV",
    "LZ_angle",
    "RAzi_angle",
    "SZ_angle",
    "ScanTime_UTC",
    "ScanTime_second",
    "Orb_mode",
    "ScanTime_year",
    "ScanTime_doy",
    "ScanTime_month",
    "ScanTime_dom",
    "ScanTime_hour",
    "ScanTime_minute",
]

interface = "readers"
family = "standard"
name = "amsua_mhs_mirs"
source_names = ["amsu-a_mhs"]

# Basic information on AMSU-B product file
# index: 1     2    3    4      5      6    7
# freq: 23.8,31.4,50.3,52.799,53.595,54.4,54.941,
# 55.499,57.29,57.29,57.29,57.29,57.29,57.29,
# index:   8     9      10    11   12   13     14
# freq: 55.499,57.29,57.29,57.29,57.29,57.29,57.29,
# index: 15  16  17    18      19     20
# freq: 89.,89.,157.,183.311,183.311,190.311

AMSUA_CHANS = [f"AMSUA_CHAN{x+1}" for x in range(15)]
MHS_CHANS = [f"Chan{x+1}_AT" for x in range(5)]

CHAN_INDEX_MAP = {x: i for i, x in enumerate(AMSUA_CHANS + MHS_CHANS)}


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read AMSU/MHS MIRS data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    ingested_xarrays = defaultdict(list)
    for fname in fnames:
        LOG.info("Reading file %s", fname)

        #     check for right input AMSU-B/MHS data file
        data_name = os.path.basename(fname).split("_")[-1].split(".")[-1]

        if data_name != "nc":
            LOG.info("Warning: wrong AMSU-B/MHS data type:  data_type=", data_name)
            raise

        if "NPR-MIRS-IMG" in os.path.basename(fname):
            LOG.info("found a NOAA MIRS AMSU-B/MHS file")
        else:
            LOG.info("not a NOAA MIRS AMSU-B/MHS file: skip it")
            raise

        """    ------  Notes  ------
        Read AMSU-B hdf files for 5 chan antenna temperature (AT) and asscoaited EDRs
            Then, transform these ATs and fields into xarray framework for GEOIPS
            ( AT will be corrected into brightness temperature (TB) later)

        Input Parameters:
            fname (str): input file name.  require a full path
        Returns:
            xarray.Dataset with required Variables and Attributes:
                Variables:
                            AMSUB vars:
                            'latitude', 'longitude', 'Ch1', 'Ch2', 'Ch3', 'Ch4','Ch4',
                            'RR', 'Snow','SWE','IWP','SFR' 'sfcType', 'time_scan'
                Attibutes:
                            'source_name', 'platform_name', 'data_provider',
                            'interpolation_radius_of_influence', 'start_datetime',
                            'end_datetime'
        """

        # import all data contents
        fileobj = h5py.File(fname, mode="r")

        var_all = {}

        # read-in selected variables
        for var in VARLIST:
            var_all[var] = fileobj[var]

        # values adjustment with scale_factor if there is a "scale_factor'

        for var in VARLIST:
            var_attrs = var_all[var].attrs.items()  # get attribute info
            for attrs in var_attrs:  # loop attribute tuples
                if "scale_factor" in attrs[0]:
                    factor = attrs[1][0]
                    var_all[var] = var_all[var] * factor

        # start_time=os.path.basename(fname).split('_')[3][1:13]
        # end_time=os.path.basename(fname).split('_')[4][1:13]
        # Collect start_time and end_time of input file (one orbit) from
        # attributes - filename wrong!
        start_datetime = datetime.strptime(
            fileobj.attrs["time_coverage_start"].astype(str), "%Y-%m-%dT%H:%M:%SZ"
        )
        end_datetime = datetime.strptime(
            fileobj.attrs["time_coverage_end"].astype(str), "%Y-%m-%dT%H:%M:%SZ"
        )

        #  -------- Apply the GEOIPS framework in XARRAY data frame ----------

        LOG.info("Making full dataframe")

        # setup the time in datetime64 format
        npix = var_all["RR"].shape[1]  # pixels per scan
        nscan = var_all["RR"].shape[0]  # total scans of this file

        time_scan = np.zeros((nscan, npix))
        # take time of each scan
        for i in range(nscan):
            yr = var_all["ScanTime_year"][i]
            dy = var_all["ScanTime_doy"][i]
            hr = var_all["ScanTime_hour"][i]
            mn = var_all["ScanTime_minute"][i]
            try:
                time_scan[i:] = "%04d%03d%02d%02d" % (yr, dy, hr, mn)
            except ValueError:
                LOG.info(
                    "Could not parse time for scan line %s: YEAR=%s, DOY=%s, "
                    "HOUR=%s, MINUTE=%s",
                    i,
                    yr,
                    dy,
                    hr,
                    mn,
                )
                continue
        #          ------  setup xarray variables   ------

        # namelist_amsub  = ['latitude', 'longitude', 'Chan1_AT', 'Chan2_AT', 'Chan3_AT'
        #                    ,'Chan4_AT','Chan5_AT', 'RR','Snow','IWP','SWE','SFR',
        #                    'Sfc_type','time']

        xarray_metadata = xr.Dataset()

        # setup attributes

        # satID and start_end time from input filename (ma3 --> METOP-c?) - TBD:
        # depending on data file )
        sat_id = os.path.basename(fname).split("_")[2]
        if sat_id == "n19":
            satid = "noaa-19"
            crtm_name = "mhs_n19"
        elif sat_id == "n18":
            satid = "noaa-18"
            crtm_name = "mhs_n18"
        elif sat_id == "n20":
            satid = "noaa-20"
            crtm_name = "mhs_n20"
        elif sat_id == "ma1":
            satid = "metop-b"
            crtm_name = "mhs_metob-b"
        elif sat_id == "ma2":
            satid = "metop-a"
            crtm_name = "mhs_metop-a"
        elif sat_id == "ma3":
            satid = "metop-c"
            crtm_name = "mhs_metop-c"

        # add attributes to xarray
        xarray_metadata.attrs["start_datetime"] = start_datetime
        xarray_metadata.attrs["end_datetime"] = end_datetime
        xarray_metadata.attrs["source_name"] = "amsu-a_mhs"
        xarray_metadata.attrs["platform_name"] = satid
        xarray_metadata.attrs["crtm_name"] = crtm_name
        xarray_metadata.attrs["data_provider"] = "NOAA-nesdis"

        # MTIFs need to be "prettier" for PMW products, so 2km resolution for
        # final image.
        # xarray_amsub.attrs['sample_distance_km'] = 15
        xarray_metadata.attrs["sample_distance_km"] = 2
        xarray_metadata.attrs["interpolation_radius_of_influence"] = 30000

        if metadata_only:
            LOG.info("metadata_only requested, returning without reading data")
            return {"METADATA": xarray_metadata}

            # Select 5 AMSU-B/MHS Channels
        xarray_amsua = xr.Dataset()
        xarray_mhs = xr.Dataset()
        xarray_amsua_mhs = xr.Dataset()
        for chan in chans:
            if chan in AMSUA_CHANS or chan in MHS_CHANS:
                chan_ind = CHAN_INDEX_MAP[chan]
                chan_num = chan_ind + 1
                if chan in AMSUA_CHANS:
                    xarray_amsua[chan] = xr.DataArray(
                        var_all["BT"][:, :, chan_ind],
                        attrs={"channel_number": chan_num},
                    )
                    xarray_amsua.attrs["source_name_override"] = "amsu-a"
                else:
                    xarray_mhs[chan] = xr.DataArray(
                        var_all["BT"][:, :, chan_ind],
                        attrs={"channel_number": chan_num},
                    )
                    xarray_mhs.attrs["source_name_override"] = "mhs"
            else:
                xarray_amsua_mhs[chan] = xr.DataArray(var_all[chan][()])
                xarray_amsua_mhs.attrs["source_name_override"] = "amsu-a_mhs"

        # Add time and geolocation variables
        beam_pos = np.broadcast_to(
            np.arange(fileobj["Field_of_view"].size) + 1, var_all["LZ_angle"][()].shape
        )
        scan_angle = xr.DataArray((beam_pos - 45.5) * 10.0 / 9.0)
        for ingest_xr in [xarray_amsua, xarray_mhs, xarray_amsua_mhs]:
            if not ingest_xr:
                continue
            source_name_override = ingest_xr.attrs["source_name_override"]
            ingest_xr.attrs = xarray_metadata.attrs
            ingest_xr.attrs["source_name"] = source_name_override
            ingest_xr["time"] = xr.DataArray(
                pd.DataFrame(time_scan)
                .astype(int)
                .apply(pd.to_datetime, format="%Y%j%H%M")
            )
            ingest_xr["latitude"] = xr.DataArray(var_all["Latitude"][()])
            ingest_xr["longitude"] = xr.DataArray(var_all["Longitude"][()])
            ingest_xr["satellite_zenith_angle"] = xr.DataArray(var_all["LZ_angle"][()])
            ingest_xr["SZ_angle"] = xr.DataArray(var_all["SZ_angle"][()])
            ingest_xr["RAzi_angle"] = xr.DataArray(var_all["RAzi_angle"][()])
            ingest_xr["sensor_scan_angle"] = scan_angle
            # These variables were previously always read in... leaving them commented
            # out as an reminder for what is in the file....
            # ingest_xr["Snow"] = xr.DataArray(var_all["Snow"][()])
            # ingest_xr["IWP"] = xr.DataArray(var_all["IWP"][()])
            # ingest_xr["SWE"] = xr.DataArray(var_all["SWE"][()])
            # ingest_xr["SFR"] = xr.DataArray(var_all["SFR"][()])
            # ingest_xr["sfcType"] = xr.DataArray(var_all["Sfc_type"][()])
            # ingest_xr["TPW"] = xr.DataArray(var_all["TPW"][()])
            # ingest_xr["CLW"] = xr.DataArray(var_all["CLW"][()])
            # ingest_xr["RWP"] = xr.DataArray(var_all["RWP"][()])
            # ingest_xr["GWP"] = xr.DataArray(var_all["GWP"][()])
            # ingest_xr["SIce"] = xr.DataArray(var_all["SIce"][()])
            # ingest_xr["SIce_MY"] = xr.DataArray(var_all["SIce_MY"][()])
            # ingest_xr["SIce_FY"] = xr.DataArray(var_all["SIce_FY"][()])
            # ingest_xr["TSkin"] = xr.DataArray(var_all["TSkin"][()])
            # ingest_xr["SurfP"] = xr.DataArray(var_all["SurfP"][()])
            # ingest_xr["CldTop"] = xr.DataArray(var_all["CldTop"][()])
            # ingest_xr["CldBase"] = xr.DataArray(var_all["CldBase"][()])
            # ingest_xr["CldThick"] = xr.DataArray(var_all["CldThick"][()])
            # ingest_xr["WindSp"] = xr.DataArray(var_all["WindSp"][()])
            # ingest_xr["WindDir"] = xr.DataArray(var_all["WindDir"][()])
            # ingest_xr["WindU"] = xr.DataArray(var_all["WindU"][()])
            # ingest_xr["WindV"] = xr.DataArray(var_all["WindV"][()])

        fileobj.close()
        if xarray_amsua:
            ingested_xarrays["AMSUA"].append(xarray_amsua)
        if xarray_mhs:
            ingested_xarrays["MHS"].append(xarray_mhs)
        if xarray_amsua_mhs:
            ingested_xarrays["AMSUA-MHS"].append(xarray_amsua_mhs)

    xarray_dict = {"METADATA": xarray_metadata}
    for dtype, xarray_list in ingested_xarrays.items():
        if xarray_list:
            final_xarray = xr.concat(xarray_list, dim="dim_0")
            final_xarray.attrs["start_datetime"] = pd.to_datetime(
                np.min(final_xarray["time"].data)
            ).to_pydatetime()
            final_xarray.attrs["end_datetime"] = pd.to_datetime(
                np.max(final_xarray["time"].data)
            ).to_pydatetime()
            xarray_dict[dtype] = final_xarray

    return xarray_dict



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""ABI Level 2 NetCDF reader."""

# Python Standard Libraries
from datetime import datetime
import logging

# Third-Party Libraries
import netCDF4 as ncdf
from numpy import isnan
import satpy
import xarray as xr

log = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "abi_l2_netcdf"
source_names = ["abi"]


def get_metadata(fname):
    """Get metadata."""
    from geoips.plugins.modules.readers.abi_netcdf import (
        _get_metadata as get_metadata,
    )

    with ncdf.Dataset(fname, "r") as df:
        metadata = get_metadata(df, fname)
    return metadata


def calculate_abi_geolocation(metadata, area_def):
    """Calculate ABI geolocation."""
    from geoips.plugins.modules.readers import abi_netcdf

    geometa = abi_netcdf._get_geolocation_metadata(metadata)
    sdt = datetime.strptime(
        metadata["file_info"]["time_coverage_start"], "%Y-%m-%dT%H:%M:%S.%fZ"
    )
    fldk_lats, fldk_lons = abi_netcdf.get_latitude_longitude(
        geometa, abi_netcdf.BADVALS, area_def
    )
    geo = abi_netcdf.get_geolocation(
        sdt, geometa, fldk_lats, fldk_lons, abi_netcdf.BADVALS, area_def
    )
    geo["fldk_lats"] = fldk_lats
    geo["fldk_lons"] = fldk_lons
    return geo


def satpy_read(fname, chans):
    """Ingest data using satpy.Scene.

    satpy.Scene object converted to xarray dataset at the end

    Parameters
    ----------
    fname : str
        Full file path
    chans : list, optional
        List of variables to load from file, by default None
        All variables are loaded in None

    Returns
    -------
    xarray.dataset
        Dataset holding either requested or all variables from file
    """
    scene = satpy.Scene([fname], reader="abi_l2_nc")
    available_vars = scene.available_dataset_names()
    if not chans:
        load_chans = available_vars
    else:
        load_chans = set(chans).intersection(set(available_vars))
    scene.load(load_chans)
    xarray = scene.to_xarray_dataset()
    return xarray


def xr_read(fname, chans=None):
    """Ingest data using xarray.open_dataset.

    Parameters
    ----------
    fname : str
        Full file path
    chans : list, optional
        List of variables to load from file, by default None
        All variables are loaded in None

    Returns
    -------
    xarray.dataset
        Dataset holding either requested or all variables from file.
    """
    ds = xr.open_dataset(fname)
    ds.attrs["start_datetime"] = datetime.strptime(
        ds.attrs["time_coverage_start"], "%Y-%m-%dT%H:%M:%S.%fZ"
    )
    ds.attrs["end_datetime"] = datetime.strptime(
        ds.attrs["time_coverage_end"], "%Y-%m-%dT%H:%M:%S.%fZ"
    )
    if chans:
        ds_vars = list(ds.variables)
        drop_vars = [x for x in ds_vars if x not in chans and x not in ["lat", "lon"]]
        ds = ds.drop_vars(drop_vars)
    return ds


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """
    Read ABI Level 2 NetCDF data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # Start with pulling metadata from the first and last files
    try:
        metadata = get_metadata(fnames[0])
        end_metadata = get_metadata(fnames[-1])
    except KeyError:
        metadata = {"file_info": xr.open_dataset(fnames[0]).attrs}
        end_metadata = {"file_info": xr.open_dataset(fnames[-1]).attrs}
    geoips_attrs = {
        "area_definition": area_def,
        "start_datetime": datetime.strptime(
            metadata["file_info"]["time_coverage_start"], "%Y-%m-%dT%H:%M:%S.%fZ"
        ),
        "end_datetime": datetime.strptime(
            end_metadata["file_info"]["time_coverage_end"], "%Y-%m-%dT%H:%M:%S.%fZ"
        ),
        "vertical_data_type": "surface",
        "source_name": "abi",
        "data_provider": "noaa",
        "interpolation_radius_of_influence": 50000,
    }
    if (
        "platform_ID" in metadata["file_info"]
        and "platform_name" not in metadata["file_info"]
    ):
        # Some of the level2 data have platform_ID attribute in the form G<NN>,
        # where <NN> = 16, 17, 18
        # Use platform_ID to set the platform_name
        geoips_attrs["platform_name"] = metadata["file_info"]["platform_ID"].replace(
            "G", "goes-"
        )

    if area_def:
        geoips_attrs["area_id"] = area_def.area_id
    else:
        geoips_attrs["area_id"] = metadata["file_info"]["scene_id"]
    meta_dataset = xr.Dataset(attrs=dict(metadata, **geoips_attrs))
    if metadata_only:
        return {"METADATA": meta_dataset}

    # C.P.C. - 20 May 2021
    # Can't figure out how to calculate the geolocation  using satpy,
    # so relying on the abi_netcdf reader for the time being until I figure it out
    # Deal with geolocation outside of loop to avoid hitting the disk numerous times
    try:
        geo = calculate_abi_geolocation(metadata, area_def)
        if area_def:
            lats = geo["latitude"]
            lons = geo["longitude"]
        else:
            lats = geo["fldk_lats"]
            lons = geo["fldk_lons"]
        ll_mask = (isnan(lats)) | (lats < -90) | (lats > 90)
    except KeyError:
        ll_mask = None

    xarrays = []
    # Ensure start_datetime and end_datetime are always tied to the file being read
    geoips_attrs.pop("start_datetime")
    geoips_attrs.pop("end_datetime")
    for fname in fnames:
        log.info("Reading %s" % fname)
        try:
            log.info("Trying with satpy backend")
            xarray = satpy_read(fname, chans)
        except ValueError:
            log.info("Attempting to load with xarray")
            xarray = xr_read(fname, chans)
        if not list(xarray.variables):
            log.info("Requested chans not in file, skipping")
            continue
        if area_def and ll_mask is not None:
            coords = dict(xarray.coords)
            coords["x"] = range(area_def.x_size)
            coords["y"] = range(area_def.y_size)
            area_dataset = xr.Dataset(coords=coords)
            area_dataset.attrs = xarray.attrs
            lines = geo["Lines"]
            samps = geo["Samples"]
            for key in xarray.keys():
                array = xarray[key].values[lines, samps]
                area_dataset[key] = (("y", "x"), array)
            xarray = area_dataset
        if ll_mask is not None:
            xarray["latitude"] = (("y", "x"), lats)
            xarray["longitude"] = (("y", "x"), lons)
            for ll in ["latitude", "longitude"]:
                xarray[ll] = xarray[ll].where(~ll_mask)
        if "lat" in xarray:
            xarray = xarray.rename({"lat": "latitude", "lon": "longitude"})
        # Add metadata needed by downstream GeoIPS utils
        xarray.attrs = dict(xarray.attrs, **geoips_attrs)
        if "start_datetime" not in xarray.attrs:
            xarray.attrs["start_datetime"] = xarray.start_time
            xarray.attrs["end_datetime"] = xarray.end_time
        xarrays.append(xarray)
    # If more than one file is passed, assuming more than one scan was passed.
    # This might not work if multiple L2 product types are passed that are of
    # different resolutions
    if len(xarrays) > 1:
        start_times = [x.attrs["start_datetime"] for x in xarrays]
        if len(set(start_times)) > 1:
            # multiple scan times were passed in
            xarray_dset = xr.concat(xarrays, dim="time_dim")
            xarray_dset.attrs["start_datetime"] = min(start_times)
            xarray_dset = xarray_dset.assign_coords({"time_dim": start_times})
        else:
            # Multiple files for a single scan time were passed in (such as multiple
            # Derived motion winds (DMW) files)
            # We can easily concatenate if these are 1D data, and have the same dim.
            # Check if we can do this, raise an error if not.
            # First check the dim sizes for each variable in the xarray
            xarray_dim_sizes = [len(dset.dims) for dset in xarrays]
            # Now check if each variable
            xarray_dim_names = ["_".join(list(dset.dims)) for dset in xarrays]
            if all([size == 1 for size in xarray_dim_sizes]) and (
                len(set(xarray_dim_names)) == 1
            ):
                xarray_dset = xr.concat(xarrays, dim=xarray_dim_names[0])
            else:
                raise ValueError(
                    "Do not know how to stack xarrays with "
                    "mismatching dim sizes and names"
                )
    else:
        xarray_dset = xarrays[0]

    return {"abi_l2_data": xarray_dset, "METADATA": meta_dataset}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `list scripts` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

from glob import glob
from importlib import resources
from os.path import basename
import pytest

from geoips.geoips_utils import is_editable

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsListScripts(BaseCliTest):
    """Unit Testing Class for List Scripts Command."""

    @property
    def command_combinations(self):
        """A list of every possible call signature for the GeoipsListScripts command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            base_args = ["geoips", "list", "scripts"]
            alias_args = ["geoips", "ls", "scripts"]
            self._cmd_list = []
            for argset in [base_args, alias_args]:
                for pkg_name in self.plugin_package_names + ["all"]:
                    if pkg_name != "all":
                        args = argset + ["-p", pkg_name]
                    else:
                        args = argset
                    self._cmd_list.append(args)
            # Add argument list which invokes the --columns flag
            self._cmd_list.append(base_args + ["--columns", "package", "filename"])
            self._cmd_list.append(base_args + ["--columns", "filename"])
            self._cmd_list.append(base_args + ["--columns", "package"])
            self._cmd_list.append(alias_args + ["--columns", "package", "filename"])
            self._cmd_list.append(alias_args + ["--columns", "filename"])
            self._cmd_list.append(alias_args + ["--columns", "package"])
            # Add argument list to retrieve help message
            self._cmd_list.append(base_args + ["-h"])
            self._cmd_list.append(alias_args + ["-h"])
            # Add argument list with a non-existent package
            self._cmd_list.append(base_args + ["-p", "non_existent_package"])
            self._cmd_list.append(alias_args + ["-p", "non_existent_package"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips list scripts ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        editable = self.assert_non_editable_error_or_wrong_package(args, error)
        if editable:
            # An error occurred using args. Assert that args is not valid and check the
            # output of the error.
            assert args != ["geoips", "list", "scripts"] and args != [
                "geoips",
                "ls",
                "scripts",
            ]
            for pkg_name in self.plugin_package_names:
                assert args != ["geoips", "list", "scripts", "-p", pkg_name]
            assert "usage: To use, type `geoips list scripts`" in error

    def check_output(self, args, output):
        """Ensure that the 'geoips list scripts ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "-h" in args:
            assert "usage: To use, type `geoips list scripts`" in output
        elif "-p" not in args and any(
            [is_editable(pkg_name) for pkg_name in self.plugin_package_names]
        ):
            self.assert_non_editable_error_or_wrong_package(args, output)
        else:
            # Checking tabular output from the list-scripts command
            if "-p" in args:
                # A certain package was requested, generate a list of all scripts from
                # that package.
                pkg_names = [args[-1]]
            else:
                # no `-p` flag, check all packages instead.
                pkg_names = self.plugin_package_names
            script_names = []
            for pkg_name in pkg_names:
                script_names += sorted(
                    [
                        basename(fpath)
                        for fpath in glob(
                            str(resources.files(pkg_name) / "../tests/scripts" / "*.sh")
                        )
                    ]
                )
            if "--columns" not in args or "filenames" in args:
                # ensure that filenames have been outputted if we are going to test this
                for script_name in script_names:
                    assert script_name in "".join(
                        output.strip().replace("\n", "").replace("│", "").split()
                    )
            selected_cols = self.retrieve_selected_columns(args)
            # Assert that the correct headers exist in the CLI output
            headers = {"GeoIPS Package": "package", "Filename": "filename"}
            self.assert_correct_headers_in_output(output, headers, selected_cols)


test_sub_cmd = TestGeoipsListScripts()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips list scripts ...' commands.

    This test covers every valid combination of commands for the 'geoips list scripts'
    command. We also test invalid commands, to ensure that the proper help documentation
    is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'list', 'scripts'])
    """
    # if "-p" not in args and
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for standard "single_channel" algorithm.

Generalized algorithm to apply data manipulation steps in a standard order
to apply corrections to a single channel output product.
"""
import logging

LOG = logging.getLogger(__name__)

KtoC_conversion = -273.15

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "single_channel"


def call(
    arrays,
    output_data_range=None,
    input_units=None,
    output_units=None,
    min_outbounds="crop",
    max_outbounds="crop",
    norm=False,
    inverse=False,
    sun_zen_correction=False,
    mask_night=False,
    max_day_zen=None,
    mask_day=False,
    min_night_zen=None,
    gamma_list=None,
    scale_factor=None,
    satellite_zenith_angle_cutoff=None,
):
    """Apply data range and requested corrections to a single channel product.

    Data manipulation steps for applying a data range and requested corrections
    to a single channel product

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data
        * MUST be length one for single_channel algorithm.
        * If applying solar zenith corrections and diurnal masking, the second index in
          the arrays list must be the solar zenith angle array. This is controlled in
          product YAML.
        * If applying a satellite zenith angle cutoff, the final index in the arrays
          list must be the satellite zenith angle array. This is controlled in the
          product YAML.
    output_data_range : list of float, default=None
        * list of min and max value for output data product.
        * This is applied LAST after all other corrections/adjustments
        * If None, use data min and max.
    input_units : str, default=None
        * Units of input data, for applying necessary conversions
        * If None, no conversion
    output_units : str, default=None
        * Units of output data, for applying necessary conversions
        * If None, no conversion
    min_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    max_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    norm : bool, default=False
        * Boolean flag indicating whether to normalize (True) or not (False)

            * If True, returned data will be in the range from 0 to 1
            * If False, returned data will be in the range from min_val to max_val
    inverse : bool, default=False
        * Boolean flag indicating whether to inverse (True) or not (False)

            * If True, returned data will be inverted
            * If False, returned data will not be inverted
    sun_zenith_correction : bool, default=False
        * Boolean flag indicating whether to apply solar zenith correction
          (True) or not (False)

            * If True, returned data will have solar zenith correction applied
              (see data_manipulations.corrections.apply_solar_zenith_correction)
            * If False, returned data will not be modified based on solar zenith
              angle
    satellite_zenith_angle_cutoff : float, default=None
        * Cutoff for masking data where satellite zenith angle exceeds threshold
        * If None, no masking

    Notes
    -----
    Order of operations, based on the passed arguments, is:

    1. Mask night
    2. Mask day
    3. Apply satellite zenith angle cutoff
    4. Apply solar zenith correction
    5. Apply gamma values
    6. Apply scale factor
    7. Convert units
    8. Apply data range.

    NOTE: If "norm=True" is specified, the "output_data_range" will NOT
    match the actual range of the returned data, since the normalized data
    will be returned between 0 and 1.

    If you require a different order of operations than that specified within
    "single_channel" algorithm, please create a new algorithm for your desired
    order of operations.

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of appropriately scaled channel data,
        in units "output_units".
    """
    data = arrays[0]
    if output_data_range is None:
        output_data_range = [data.min(), data.max()]

    # Mask everything greater than max_day_zen
    # day zenith angles are less than 90
    # night zenith angles are greater than 90
    if max_day_zen and len(arrays) >= 2:
        from geoips.data_manipulations.info import percent_unmasked
        from geoips.data_manipulations.corrections import mask_night

        sun_zenith = arrays[1]
        LOG.info("Percent unmasked day/night %s", percent_unmasked(data))
        data = mask_night(data, sun_zenith, max_day_zen)
        LOG.info("Percent unmasked day only %s", percent_unmasked(data))

    # Mask everything less than min_night_zen
    # day zenith angles are less than 90
    # night zenith angles are greater than 90
    if mask_day and min_night_zen and len(arrays) >= 2:
        from geoips.data_manipulations.info import percent_unmasked
        from geoips.data_manipulations.corrections import mask_day

        sun_zenith = arrays[1]
        LOG.info("Percent unmasked day/night %s", percent_unmasked(data))
        data = mask_day(data, sun_zenith, min_night_zen)
        LOG.info("Percent unmasked night only %s", percent_unmasked(data))

    if satellite_zenith_angle_cutoff is not None and len(arrays) >= 2:
        from geoips.data_manipulations.info import percent_unmasked
        from geoips.data_manipulations.corrections import apply_satellite_zenith_cutoff

        sat_zenith = arrays[-1]
        data = apply_satellite_zenith_cutoff(
            data, sat_zenith, satellite_zenith_angle_cutoff
        )
        LOG.info(
            "Percent unmasked after zenith angle cutoff %s", percent_unmasked(data)
        )

    if sun_zen_correction and len(arrays) >= 2:
        sun_zenith = arrays[1]
        from geoips.data_manipulations.corrections import apply_solar_zenith_correction

        data = apply_solar_zenith_correction(data, sun_zenith)

    if gamma_list is not None:
        from geoips.data_manipulations.corrections import apply_gamma

        for gamma in gamma_list:
            data = apply_gamma(data, gamma)

    if scale_factor is not None:
        from geoips.data_manipulations.corrections import apply_scale_factor

        data = apply_scale_factor(data, scale_factor)

    from geoips.data_manipulations.conversions import unit_conversion

    data = unit_conversion(data, input_units, output_units)

    from geoips.data_manipulations.corrections import apply_data_range

    data = apply_data_range(
        data,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )

    return data



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Output Checkers interface module."""

from geoips.interfaces.base import (
    BaseModuleInterface,
    BaseModulePlugin,
    ValidationError,
)
import logging

# import subprocess
from geoips.commandline.log_setup import log_with_emphasis
import gzip
from glob import glob
from os.path import exists, splitext, basename, dirname, isdir, join
from os import makedirs, getenv
from shutil import copyfileobj
from geoips.filenames.base_paths import make_dirs


LOG = logging.getLogger(__name__)
rezip = False


def write_bad_comparisons_to_file(badcomps, compare_products, compare_strings, diffdir):
    """Write text file with cp commands to update products with bad comparisons.

    Write a text file that can be sourced to replace test output comparison files
    with the most recent outputs when the output checker returns a bad comparisons.

    When GeoIPS is updated in a way that changes the imagery and those changes
    are expected, the integration test image comparisons will fail. This
    function writes a file named `cp_BADCOMPARES.txt` which can be
    sourced to copy the new comparison files into place, overwriting
    the previous comparison files. This allows easily updating the test
    comparison outputs when code updates make expected changes to the
    output imagery.

    We must ensure we also gzip the comparison product in the test outputs
    directory after copying if the original test comparison output file was
    gzipped.  If applicable, a gzip command will also be added to
    cp_BADCOMPARES.txt.

    Parameters
    ----------
    badcomps: list
        List of files that returned a "BAD COMPARISON" from the output checker.
        These are the full paths to the files produced during the most recent
        run.
    compare_products: dict
        dictionary of comparison products - including the original comparison
        product file name, the gunzipped product file path if applicable,
        and the path to the file that should be used for comparison against
        "badcomps" files.
    compare_strings : list of str
        List of all comparison "tags" included in goodcomps and badcomps lists.

        * This list is used to remove the comparison tags from goodcomps and
          badcomps to retrieve only the file path.
    diffdir: str
        Full path to write "cp_BADCOMPARES.txt" file.
        NOTE: diffdir currently also is used to specify the path to the
        output files. Eventually we should pass in 2 separate paths, one
        to specify where to write the diff files, and a separate path
        indicating where to write the updated test output comparison files.

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully, non-zero
        if there were bad comparisons.
    """
    for badcompare in badcomps:
        LOG.warning("BADCOMPARE %s", badcompare)
    if len(badcomps) > 0:
        fname_cp = join(diffdir, "cp_BADCOMPARES.txt")
        fname_badcptest = join(diffdir, "cptest_BADCOMPARES.txt")
        LOG.interactive(
            "BADCOMPARES Commands to copy %d files that had bad comparisons",
            len(badcomps),
        )
        LOG.debug(f"  source {fname_badcptest}")
        LOG.interactive(f"  source {fname_cp}")
        # Include print statement for easy copy/paste at the command line
        print(f"***\nsource {fname_cp}\n***")
        comparison_path = join(diffdir, "..")
        test_path = join(diffdir, "BADCOMPARES")
        with open(fname_cp, "a") as fobj:
            for badcomp in badcomps:
                if compare_strings is not None:
                    for compare_string in compare_strings:
                        badcomp = badcomp.replace(compare_string, "")
                fobj.write(f"cp -v {badcomp} {comparison_path}\n")
                comparison_basename = basename(badcomp)
                comparison_filename = join(comparison_path, comparison_basename)
                comparison_gz_filename = comparison_filename + ".gz"
                if (
                    basename(comparison_gz_filename) in compare_products
                    and compare_products[basename(comparison_gz_filename)][
                        "gunzipped_comparison"
                    ]
                ):
                    fobj.write(f"gzip {comparison_filename}\n")
        with open(fname_badcptest, "a") as fobj:
            fobj.write(f"mkdir {test_path}\n")
            for badcomp in badcomps:
                badcomp = badcomp.replace("IMAGE ", "")
                badcomp = badcomp.replace("TEXT ", "")
                badcomp = badcomp.replace("GEOIPS NETCDF ", "")
                badcomp = badcomp.replace("GEOTIFF ", "")
                # For display purposes - tifs are easier to view
                test_basename = basename(badcomp).replace(".jif", ".tif")
                test_filename = join(test_path, test_basename)
                fobj.write(f"cp {badcomp} {test_filename}\n")
    # Assemble the return value and return.
    # Bad comparisons are bits 0/1 of the return code.
    retval = 0
    if len(badcomps) != 0:
        retval = len(badcomps)
        if len(badcomps) > 4:
            retval = 4
        LOG.info("BADCOMPS %s", len(badcomps))
    return retval << 0


def write_remove_temp_files_to_file(remove_temp_files, diffdir):
    """Write text file with rm commands to remove temporary files.

    Write a text file ``rm_TEMPFILES.txt`` that can be sourced to remove
    temporary files that were created in the ``$GEOIPS_OUTDIRS`` scratch
    directory during the output comparison checks during integration tests.
    This currently includes temporary gunzipped versions of products
    (ie, if a product was gzipped during processing, then gunzipped to
    a temporary location during the comparison tests, it will get added
    to the `remove_temp_files` list), and may include additional temporary
    files in the future.

    The gunzipped files are not auto-deleted during the comparison process,
    because they must be manually copied to the tets outputs directory after
    processing is complete, so we do not remove the files at runtime.

    Parameters
    ----------
    remove_temp_files: list
        List of files that should be written to ``rm_TEMPFILES.txt`` for deletion.
        This currently includes auto-gunzipped products, and potentially
        additional temporary files in the future.
    diffdir: str
        Full path to write ``rm_TEMPFILES.txt`` file
        NOTE: `diffdir` currently also is used to specify the path to the
        output files. Eventually we should pass in 2 separate paths, one
        to specify where to write the diff files, and a separate path
        indicating where to write the updated test output comparison files.

    Returns
    -------
    int
        Binary code: 0 for successfully written ``rm_TEMPFILES.txt`` file.
    """
    if len(remove_temp_files) > 0:
        fname_rmtemp = join(diffdir, "rm_TEMPFILES.txt")
        LOG.interactive(
            "RMTEMPFILES Commands to remove %d "
            "temp files from gunzipped files scratch path.",
            len(remove_temp_files),
        )
        LOG.interactive(f"  source {fname_rmtemp}")
        # Include print statement for easy copy/paste at the command line
        print(f"***\nsource {fname_rmtemp}\n***")
        with open(fname_rmtemp, "a") as fobj:
            for remove_temp_file in remove_temp_files:
                fobj.write(f"rm -v {remove_temp_file}\n")
    return 0


def write_missing_products_to_file(missingproducts, compare_products, diffdir):
    """Write text file with rm commands to remove products from comparison directory.

    Write a text file contatining rm commands that can be sourced to remove
    existing test output comparison files from the test comparison directories
    that were NOT found in the most recent output.

    When GeoIPS is updated in a way that intentionally causes a previously
    expected output to no longer be produced, it will cause the integration
    test comparison to fail with a MISSING PRODUCT error (ie, a file
    exists in the test comparison directory, but is not a current output
    product).

    When this is intended behaviour due to recent code updates, we must
    remove that product from the test output comparison directories.

    This function writes a file named `rm_MISSINGPRODUCTS.txt` which
    can be sourced to remove the files from the test comparison directory
    that were NOT found in the list of current output products. This
    allows easily updating the test comparison outputs when code updates
    make expected changes to the output products.

    Parameters
    ----------
    missingproducts: list
        List of files that returned a "MISSING PRODUCT" from the output checker.
        These are the full paths to the files produced during the most recent
        run.
    compare_products: dict
        dictionary of comparison products - including the original comparison
        product file name, the gunzipped product file path if applicable,
        and the path to the file that should be used for comparison against
        "missingproducts" files.
    diffdir: str
        Full path to write "rm_MISSINGPRODUCTS.txt" file
        NOTE: diffdir currently also is used to specify the path to the
        output files. Eventually we should pass in 2 separate paths, one
        to specify where to write the diff files, and a separate path
        indicating where to write the updated test output comparison files.

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully, non-zero
        if there were missing products.
    """
    # This should be in loop below
    # for compareproduct in compare_products:
    #     LOG.warning(
    #         "MISSINGPRODUCT not in current run: %s",
    #         basename(compareproduct),
    #     )
    # If we have any missingproducts, loop through in order to write them to
    # file for manual removal from the test comparison directories.
    if len(missingproducts) > 0:
        # Actual rm commands to remove files
        fname_rm = join(diffdir, "rm_MISSINGPRODUCTS.txt")
        # cp commands to copy over the products for review to a temporary
        # directory.  This does not impact the actual test outputs.
        fname_missingprodcptest = join(diffdir, "cptest_MISSINGPRODUCTS.txt")
        LOG.interactive(
            "MISSINGPRODUCTS Commands to remove %d "
            "incorrect files from test output path.",
            len(missingproducts),
        )
        LOG.debug("  source {0}".format(fname_missingprodcptest))
        # If there are missing products, ensure the source command to
        # remove them from the comparison test outputs directory is
        # printed at the interactive log level
        LOG.interactive("  source {0}".format(fname_rm))
        # Include print statement for easy copy/paste at the command line
        print("***\nsource {0}\n***".format(fname_rm))
        test_path = join(diffdir, "MISSINGPRODUCTS")
        with open(fname_rm, "a") as fobj:
            # Now loop through each missing product to write them to the file
            for missingproduct in missingproducts:
                missingproduct_basename = basename(missingproduct)
                for compare_product in compare_products:
                    # Pull the file that is actually used for comparison from
                    # the compare_products dictionary.
                    file_for_comparison = compare_products[compare_product][
                        "file_for_comparison"
                    ]
                    file_for_comparison_basename = basename(file_for_comparison)
                    # If the basename of the current missing product matches the
                    # basename of the file for comparison, then write out the
                    # command to remove the actual stored comparison file from
                    # the test repo.
                    if missingproduct_basename == file_for_comparison_basename:
                        comparison_filename = compare_products[compare_product][
                            "stored_comparison"
                        ]
                        fobj.write(f"  rm -v {comparison_filename}\n")
                        LOG.debug(f"    TEST OUTPUT: {comparison_filename}")
                        LOG.warning(
                            "MISSINGPRODUCT not in current run: %s",
                            basename(comparison_filename),
                        )
                    else:
                        comparison_filename = compare_products[compare_product][
                            "stored_comparison"
                        ]
                        fobj.write(f"  rm -v {comparison_filename}\n")
                        LOG.debug(f"    TEST OUTPUT: {file_for_comparison}")
                        LOG.warning(
                            "MISSINGPRODUCT not in current run: %s",
                            basename(comparison_filename),
                        )
        # This is just for testing purposes - write out copy commands to
        # copy the missing product into a test location for easy review.
        with open(fname_missingprodcptest, "a") as fobj:
            fobj.write(f"mkdir {test_path}\n")
            for missingproduct in missingproducts:
                # For display purposes - tifs are easier to view
                test_basename = basename(missingproduct).replace(".jif", ".tif")
                test_filename = join(test_path, test_basename)
                fobj.write(f"cp -v {missingproduct} {test_filename}\n")
    # Assemble the return value and return.
    # Missing products are bits 4/5 of the return code.
    retval = 0
    if len(missingproducts) != 0:
        retval = len(missingproducts)
        if len(missingproducts) > 4:
            retval = 4
        LOG.info("MISSINGPRODUCTS %s", len(missingproducts))
    return retval << 4


def write_missing_comparisons_to_file(missingcomps, diffdir):
    """Write text file with cp commands to add products to comparison directory.

    Write a text file contatining cp commands that can be sourced to add
    new test output comparison files to the test comparison directories
    that were found in the most recent output and were not previously in
    the test comparison directory.

    When GeoIPS is updated in a way that intentionally causes a new
    output product, it will cause the integration test comparison to fail
    with a MISSING COMPARE error (ie, a file exists in the current output
    product path, but does not exist in the test comparison directory).

    When this is intended behaviour due to recent code updates, we must
    add that new product to the test output comparison directories from
    the current output path.

    This function writes a file named `cp_MISSINGCOMPARE.txt` which
    can be sourced to copy the appropriate files from the current
    output product path into the test comparison directory. This
    allows easily updating the test comparison outputs when code updates
    make expected changes to the output products.

    Parameters
    ----------
    missingcomps: list
        List of files that returned a "MISSING COMPARE" from the output checker.
        These are the full paths to the files produced during the most recent
        run.
    diffdir: str
        Full path to write "rm_MISSINGCOMPARE.txt" file
        NOTE: diffdir currently also is used to specify the path to the
        output files. Eventually we should pass in 2 separate paths, one
        to specify where to write the diff files, and a separate path
        indicating where to write the updated test output comparison files.

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully, non-zero
        if there were missing comparisons.
    """
    for missingcompare in missingcomps:
        LOG.warning("MISSINGCOMPARE not in comparepath: %s", basename(missingcompare))
    if len(missingcomps) > 0:
        fname_cp = join(diffdir, "cp_MISSINGCOMPARE.txt")
        fname_missingcompcptest = join(diffdir, "cptest_MISSINGCOMPARE.txt")
        LOG.interactive(
            "MISSINGCOMPARE Commands to copy %d missing files to comparison path.",
            len(missingcomps),
        )
        LOG.debug("  source {0}".format(fname_missingcompcptest))
        LOG.interactive("  source {0}".format(fname_cp))
        # Include print statement for easy copy/paste at the command line
        print("***\nsource {0}\n***".format(fname_cp))
        comparison_path = join(diffdir, "..")
        test_path = join(diffdir, "MISSINGCOMPARE")
        with open(fname_cp, "a") as fobj:
            for missingcomp in missingcomps:
                # We are passing the missingcomp directly to the diff dir
                # We have no knowledge if these should be gzipped or not,
                # so they are just copied as is.  Developers must manually
                # gzip before commiting if desired.
                fobj.write(f"  cp -v {missingcomp} {comparison_path}\n")
                LOG.debug(f"    CURR OUTPUT: {missingcomp}")
        with open(fname_missingcompcptest, "a") as fobj:
            fobj.write(f"mkdir {test_path}\n")
            for missingcomp in missingcomps:
                # For display purposes - tifs are easier to view
                test_basename = basename(missingcomp).replace(".jif", ".tif")
                test_filename = join(test_path, test_basename)
                fobj.write(f"cp -v {missingcomp} {test_filename}\n")
    # Assemble the return value and return.
    # Missing comparisons are bits 2/3 of the return code.
    retval = 0
    if len(missingcomps) != 0:
        retval = len(missingcomps)
        if len(missingcomps) > 4:
            retval = 4
        LOG.info("MISSINGCOMPS %s", len(missingcomps))
    return retval << 0


def write_good_comparisons_to_file(goodcomps, compare_strings, diffdir):
    """Write all good comparisons to a file for easy collection of outputs.

    During integration tests, if the current product output matches the
    test output stored in the test comparison directories, the output
    checkers will return 0 as a GOOD COMPARE.  These outputs do not need
    to be updated in the test comparison directories, since the current
    products and test comparison outputs match within tolerance.

    The developer may desire copying these files to a single location
    convenience, however.

    This function writes a file named `cptest_GOODCOMPARE.txt` which
    can be sourced to copy all products that were identical to the
    current output products to a common location, for convenience.
    This does not impact any files stored in the test comparison
    directories.

    Parameters
    ----------
    goodcomps: list
        List of files that returned a "GOOD COMPARE" from the output checker.
        These are the full paths to the files produced during the most recent
        run.
    compare_strings : list of str
        List of all comparison "tags" included in goodcomps and badcomps lists.

        * This list is used to remove the comparison tags from goodcomps and
          badcomps to retrieve only the file path when writing the cp commands.
    diffdir: str
        Full path to write "cptest_GOODCOMPARE.txt" file
        NOTE: diffdir currently also is used to specify the path to the
        output files. Eventually we should pass in 2 separate paths, one
        to specify where to write the diff files, and a separate path
        indicating where to write the updated test output comparison files.

    Returns
    -------
    int
        Binary code: 0 for good comparisons
    """
    for goodcompare in goodcomps:
        LOG.warning("GOODCOMPARE %s", goodcompare)
    if len(goodcomps) > 0:
        fname_goodcptest = join(diffdir, "cptest_GOODCOMPARE.txt")
        LOG.debug(f"source {fname_goodcptest}")
        test_path = join(diffdir, "GOODCOMPARE")
        with open(fname_goodcptest, "a") as fobj:
            fobj.write(f"mkdir {test_path}\n")
            for goodcomp in goodcomps:
                if compare_strings is not None:
                    for compare_string in compare_strings:
                        goodcomp = goodcomp.replace(compare_string, "")
                # For display purposes - tifs are easier to view
                test_basename = basename(goodcomp).replace(".jif", ".tif")
                test_filename = join(test_path, test_basename)
                fobj.write(f"cp {goodcomp} {test_filename}\n")
    return 0


def get_missing_products(output_products_for_comparison, compare_products):
    """Get list of missing products from compare_products_dictionary.

    Parameters
    ----------
    output_products_for_comparison: list
        List of output products from the current run.
    compare_products: dict
        Dictionary of products found in the test comparison directory.
        Note the key of the compare_products dictionary is the filename that
        is stored in the comparison directory (ie, this may be a gzipped file).
        We must use the "gunzipped_comparison", "stored_comparison", and
        "file_for_comparison" keys within that dictionary to ensure we are
        accessing the correct versions of the files.

    Returns
    -------
    missingproducts : list
        List of products that were missing from the current run.
    """
    missingproducts = []
    for output_product_for_comparison in output_products_for_comparison:
        found_one = False
        # compare_products is a dictionary,
        # with "basename(compare_product)" as the keys.
        # Add the full paths to the missing files as the missingproducts.
        for basename_compare_product in compare_products:
            # We want to check that file_for_comparison is in the list of
            # output_products_for_comparison.  We know file_for_comparison will always
            # be the file corresponding to the actual current output products.
            # gunzipped_comparisons may be None, and stored_comparisons may be
            # gzipped.
            file_for_comparison = compare_products[basename_compare_product][
                "file_for_comparison"
            ]
            if basename(file_for_comparison) == basename(output_product_for_comparison):
                found_one = True
        if not found_one:
            missingproducts += [output_product_for_comparison]
    return missingproducts


def is_gz(fname):
    """Check if fname is a gzip file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a gz file, False otherwise.
    """
    if splitext(fname)[-1] in [".gz"]:
        return True
    # If a file exists with the base fname appended with .gz,
    # then use that.
    if glob(fname + ".gz"):
        return True
    return False


def gunzip_product(fname, is_comparison_product=False, clobber=False):
    """Gunzip file fname.

    Parameters
    ----------
    fname : str
        File to gunzip.
    is_comparison_product : bool, default False
        * If True, set to $GEOIPS_OUTDIRS/gunzipped_products/comparison_products
        * If False, set to $GEOIPS_OUTDIRS/gunzipped_products/output_products
    clobber : bool, default False
        If True, overwrite existing file, else do nothing.

    Returns
    -------
    str
        Filename after gunzipping
    """
    # If it is already a .gz file, use that.
    if splitext(fname)[-1] in [".gz"]:
        gz_fname = fname
    # Otherwise, see if the .gz version exists.
    elif glob(fname + ".gz"):
        gz_fname = fname + ".gz"

    messages = ["Gunzipping product for comparisons"]
    messages.append(f"gunzip {gz_fname}")
    log_with_emphasis(LOG.info, *messages)

    if is_comparison_product:
        save_dir = join(
            getenv("GEOIPS_OUTDIRS"),
            "scratch",
            "gunzipped_products",
            "comparison_products",
        )
    else:
        save_dir = join(
            getenv("GEOIPS_OUTDIRS"), "scratch", "gunzipped_products", "output_products"
        )
    make_dirs(save_dir)
    gunzip_basename = str(basename(gz_fname)).replace(".gz", "")
    gunzip_filename = join(save_dir, gunzip_basename)
    if clobber or not exists(gunzip_filename):
        LOG.interactive("Gunzipping file %s to %s", gz_fname, save_dir)
        # Read in the compressed data
        with gzip.open(gz_fname, "rb") as f_in:
            with open(gunzip_filename, "wb") as f_out:
                copyfileobj(f_in, f_out)
    return gunzip_filename


class OutputCheckersBasePlugin(BaseModulePlugin):
    """Output Checkers Base Plugin for comparing data outputs."""

    def is_gz(self, fname):
        """Check if fname is a gzip file.

        Parameters
        ----------
        fname : str
            Name of file to check.

        Returns
        -------
        bool
            True if it is a gz file, False otherwise.
        """
        return is_gz(fname)

    def gunzip_product(self, fname, is_comparison_product=False, clobber=False):
        """Gunzip file fname.

        Parameters
        ----------
        fname : str
            File to gunzip.
        is_comparison_product : bool, default False
            * If True, set to $GEOIPS_OUTDIRS/gunzipped_products/comparison_products
            * If False, set to $GEOIPS_OUTDIRS/gunzipped_products/output_products
        clobber : bool, default False
            If True, overwrite existing file, else do nothing.

        Returns
        -------
        str
            Filename after gunzipping
        """
        return gunzip_product(fname, is_comparison_product, clobber)

    def get_out_diff_fname(self, compare_product, output_product, ext=None, flag=None):
        """Obtain the filename for output and comparison product diff.

        Parameters
        ----------
        compare_product : str
            Full path to product filename in the comparison directory
        output_product : str
            Full path to product filename in the current output directory
        ext : str, default=None
            Extension to use as an alternative to the original file extension
        flag : str, default=None
            Additional identifying string to include in output diff filename

        Returns
        -------
        out_diff_fname : str
            Full path to output diff file.
        """
        if not flag:
            flag = ""
        diffdir = join(
            dirname(compare_product), "diff_test_output_dir_{0}".format(getenv("USER"))
        )
        out_diff_fname = join(
            diffdir, "diff_test_output_" + flag + basename(output_product)
        )
        if not exists(diffdir):
            makedirs(diffdir)
        # Allow outputing a different extension than the original.
        if ext is not None:
            out_diff_fname = splitext(out_diff_fname)[0] + ext
        # Create a png diff image for simplicity
        elif splitext(out_diff_fname)[-1] == ".jif":
            out_diff_fname = splitext(out_diff_fname)[0] + ".png"
        return out_diff_fname

    def get_compare_products(self, compare_path):
        """Get dictionary of compare products from the compare_path.

        Parameters
        ----------
        compare_path: str
            Full path to the comparison directory to populate compare_products
            dictionary.

        Returns
        -------
        compare_products: dict
            Dictionary of products found in the test comparison directory.
            The key of the compare_products dictionary is the filename that
            is stored in the comparison directory (ie, this may be a gzipped file).

            * stored_comparison is the actual name stored in the comparison test dir
            * gunzipped comparison is the gunzipped version of the file, or None if
              file was not gzipped
            * file_for_comparison is "compare_product" if the file is not gzipped,
              and "gunzipped_comparison" if file is gzipped.
        remove_temp_files: list
            List of full paths to files that must be removed after processing
            (gunzipped files in GEOIPS_OUTDIRS scratch directory)
        """
        # Here we are collecting the basenames of the files for comparison, as
        # well as the associated gz filename if the original comparison product
        # was gzipped.
        compare_products = {}
        remove_temp_files = []
        # Loop through all comparison products found in "compare_path". We want
        # to collect ALL products available for comparison.
        for compare_product in glob(compare_path + "/*"):
            if isdir(compare_product):
                # Skip 'diff' output directory
                continue
            file_for_comparison = compare_product
            compare_product_gunzipped = None
            # If the current product is gzipped, we want to gunzip it for comparison.
            # We always gunzip for comparison.
            if self.is_gz(compare_product):
                # self.gunzip_product returns the resulting filename (which may be
                # in a different directory). NOTE: this is the first time gunzip
                # is run on compare_products, so ensure we clobber what is there.
                compare_product_gunzipped = self.gunzip_product(
                    compare_product, is_comparison_product=True, clobber=True
                )
                remove_temp_files += [compare_product_gunzipped]
                file_for_comparison = compare_product_gunzipped
            # This is a complete list of all comparison products.
            # stored_comparison is the actual name stored in the comparison test dir
            # gunzipped comparison is the gunzipped version of the file, or None if
            #   file was not gzipped
            # file_for_comparison is "compare_product" if the file is not gzipped,
            #   and "gunzipped_comparison" if file is gzipped.
            compare_products[basename(compare_product)] = {
                "stored_comparison": compare_product,
                "gunzipped_comparison": compare_product_gunzipped,
                "file_for_comparison": file_for_comparison,
            }

        return compare_products, remove_temp_files

    def perform_comparisons(
        self,
        output_products,
        compare_products,
        **kwargs,
    ):
        """Get the final output products list from the original list of outputs.

        Parameters
        ----------
        output_products : list
            This is the original list of output products, exactly as they are
            produced during processing.  If a product is gzipped during processing,
            the output_products list will include the gzipped file.
        compare_products : dict
            This is a dictionary containing all of the comparison products available
            in the test outputs directory.

            * Keys: basename of actual output product as created during run
              (ie, gzipped if applicable)

              * key: "file_for_comparison", value: full path to file that is used
                when comparing against the current output (ie, could be temporary
                gunzip location).
              * key: "gunzipped_comparison", value: None if original output file was
                not gzipped, full path to gunzipped file if original output file
                was gzipped.
              * key: "stored_comparison", value: full path to file that is stored in
                the tests outputs location (ie, could be gzipped)
        kwargs : dict
            keyword arguments to pass directly through to test products function.

        Returns
        -------
        output_products_for_comparison : list
            List of full paths to files that will be used for comparisons.  Ie,
            if the output product was gzipped, this would be the full path to
            the gunzipped file in the temporary location.
        missingcomps : list
            List of full paths to files that were found in the current run, and NOT
            found in the comparison directory.
        goodcomps : list
            List of full paths to files that had good comparisons.
        badcomps : list
            List of full paths to files that had bad comparisons between the
            current output products and the test comparison outputs.
        compare_strings : list
            List of strings included in log output for each type of comparison.
        remove_temp_files : list
            List of full paths to files that should be removed from temporary scratch
            directory.
        """
        output_products_for_comparison = []
        remove_temp_files = []
        missingcomps = []
        goodcomps = []
        badcomps = []
        compare_strings = []

        for output_product in output_products:
            # Skip 'diff' output directory
            if isdir(output_product):
                continue

            output_product_for_comparison = output_product
            # Gunzip the output product if it is a gzipped file.
            # We will perform all comparisons on the gunzipped form.
            if is_gz(output_product):
                output_product_for_comparison = self.gunzip_product(
                    output_product, is_comparison_product=False
                )
                remove_temp_files += [output_product_for_comparison]

            log_with_emphasis(LOG.info, f"COMPARE {output_product_for_comparison}")
            found_one = False
            for compare_product in compare_products:
                file_for_comparison = compare_products[basename(compare_product)][
                    "file_for_comparison"
                ]
                # If the current output product matches the basename of the
                # gunzipped comparison product, then we have a match and
                # we can actually compare the files.  Whether a file is
                # gzipped in the current output directory, or the comparison
                # directory, we ALWAYS compare the gunzipped version, so this
                # comparison will always hold (output_product will always be
                # gunzipped, as well gunzipped_comparison).
                if basename(output_product_for_comparison) == basename(
                    file_for_comparison
                ):
                    found_one = True
                    test_product_func = self.test_products
                    goodcomps, badcomps, compare_strings = test_product_func(
                        output_product_for_comparison,
                        file_for_comparison,
                        goodcomps,
                        badcomps,
                        compare_strings,
                        **kwargs,
                    )
            # If we didn't find an exact filename match, then we have a missing
            # comparison, so append to the missingcomps list.
            if not found_one:
                missingcomps += [output_product_for_comparison]
            output_products_for_comparison += [output_product_for_comparison]

            LOG.info("")
        return (
            output_products_for_comparison,
            missingcomps,
            goodcomps,
            badcomps,
            compare_strings,
            remove_temp_files,
        )

    def compare_outputs(
        self,
        compare_path,
        output_products,
        **kwargs,
    ):
        """Compare the "correct" imagery found the list of current output_products.

        Compares files produced in the current processing run with the list of
        "correct" files contained in "compare_path".

        Parameters
        ----------
        compare_path : str
            Path to directory of "correct" products - filenames must match
            output_products
        output_products : list of str
            List of strings of current output products,
            to compare with products in compare_path
        kwargs: dict
            Dictionary containing kwargs for comparing products.
            This gets passed through to the "test_products" method.

        Returns
        -------
        int
            Binary code: 0 if all comparisons were completed successfully.
        """
        log_with_emphasis(LOG.info, f"COMPARISONS OF KNOWN OUTPUTS IN {compare_path}")
        # We gunzip comparison files to a temporary location, so keep track of
        # all the temporary files so we can remove them at the end.
        remove_temp_files = []

        compare_products, curr_remove_temp_files = self.get_compare_products(
            compare_path
        )
        remove_temp_files += curr_remove_temp_files

        # This returns the list of output products that will be used for
        # comparisons (ie, the gunzipped version if the original was gzipped),
        # the list of missing comparison products, and appends to the list
        # of temp files to remove.
        (
            output_products_for_comparison,
            missingcomps,
            goodcomps,
            badcomps,
            compare_strings,
            curr_remove_temp_files,
        ) = self.perform_comparisons(
            output_products,
            compare_products,
            **kwargs,
        )
        remove_temp_files += curr_remove_temp_files
        log_with_emphasis(
            LOG.info,
            f"DONE RUNNING COMPARISONS OF KNOWN OUTPUTS IN {compare_path}",
        )

        # Identify all missing products based on the list of output products
        # and the dictionary of comparison products
        missingproducts = get_missing_products(
            output_products_for_comparison, compare_products
        )

        diffdir = join(compare_path, f"diff_test_output_dir_{getenv('USER')}")
        if not exists(diffdir):
            makedirs(diffdir)

        retval = 0
        retval += write_good_comparisons_to_file(goodcomps, compare_strings, diffdir)
        retval += write_missing_comparisons_to_file(missingcomps, diffdir)
        retval += write_missing_products_to_file(
            missingproducts, compare_products, diffdir
        )
        retval += write_remove_temp_files_to_file(remove_temp_files, diffdir)
        retval += write_bad_comparisons_to_file(
            badcomps, compare_products, compare_strings, diffdir
        )

        LOG.info("retval: %s", bin(retval))
        if retval != 0:
            LOG.info(
                "Nonzero return value indicates error, 6 bit binary code: " "WXY where:"
            )
            LOG.info(
                (
                    "    Y (0-1): BADCOMP: number of bad comparisons found between"
                    "comparepath and current run output path"
                )
            )
            LOG.info(
                (
                    "    X (2-3): MISSINGCOMP: Number of products missing in "
                    "compare path but existing in current output path"
                )
            )
            LOG.info(
                (
                    "    W (4-5): MISSINGPROD: Number of products existing in "
                    "comparepath, but missing in current output path"
                )
            )

        return retval

    def test_products(
        self,
        output_product,
        compare_product,
        goodcomps,
        badcomps,
        compare_strings,
        **kwargs,
    ):
        """Test output_product against "good" product stored in "compare_path".

        Parameters
        ----------
        output_product : str
            * Full path to current output product
        compare_product : str
            * Full path to "good" comparison product
        goodcomps : list of str
            * List of full paths to all "good" successful comparisons
              (output and compare images match)
            * Each str is prepended with a "compare_string" tag to identify which
              comparison type was performed.
        badcomps : list of str
            * List of full paths to all "bad" unsuccessful comparisons
              (output and compare images differ)
            * Each str is prepended with a "compare_string" tag to identify which
              comparison type was performed.
        compare_strings : list of str
            * List of all comparison "tags" included in goodcomps and badcomps lists.
            * This list is used to remove the comparison tags from goodcomps and
              badcomps to retrieve only the file path.
        kwargs: dict
            Additional arguments to pass through to "outputs_match" method.

        Returns
        -------
        goodcomps: list of str
            All current good comparisons appended to the list passed in.
        badcomps: list of str
            All current bad comparisons appended to the list passed in.
        compare_strings: list of str
            All current comparison "tags" added to the list passed in.

        Raises
        ------
        TypeError
            Raised when current output product does not have an associated
            comparison test defined.
        """
        if self.is_gz(compare_product):
            compare_product = self.gunzip_product(
                compare_product, is_comparison_product=True
            )
        if self.name == "netcdf":
            comp_str = "GEOIPS NETCDF "
        else:
            comp_str = self.name.upper() + " "
        compare_strings += [comp_str]
        if self.module.outputs_match(
            self,
            output_product,
            compare_product,
            **kwargs,
        ):
            goodcomps += [comp_str + "{0}".format(output_product)]
        else:
            badcomps += [comp_str + "{0}".format(output_product)]

        return goodcomps, badcomps, compare_strings


class OutputCheckersInterface(BaseModuleInterface):
    """Output Checkers routines to apply when comparing data outputs."""

    name = "output_checkers"
    required_args = {"standard": {}}
    required_kwargs = {"standard": {}}
    plugin_class = OutputCheckersBasePlugin
    # required_args = {
    #     "standard": ["fname", "output_product", "compare_product"],
    #     "print_gunzip": ["fobj", "gunzip_fname"],
    #     "print_gzip": ["fobj", "gzip_fname"],
    # }
    # required_kwargs = {"standard": {}}
    # allowable_kwargs = {
    #     "test_product": ["goodcomps", "badcomps", "compare_strings"],
    #     "out_diffname": ["ext", "flag"],
    #     "compare_outputs": ["test_product_func"],
    # }

    def identify_checker(self, filename):
        """Identify the correct output checker plugin and return its name."""
        checker_found = False
        checker_name = None
        if is_gz(filename):
            # NOTE: this is the first time gunzip is run on the current
            # product, so ensure we clobber what is already there.
            filename = gunzip_product(
                filename, is_comparison_product=False, clobber=True
            )
        for output_checker in self.get_plugins():
            checker_found = output_checker.module.correct_file_format(filename)
            if checker_found:
                checker_name = output_checker.module.name
                break
        if not checker_found:
            raise TypeError("There isn't an output checker built for this data type.")
        return checker_name

    def get_plugin(self, name, rebuild_registries=None):
        """Return the output checker plugin corresponding to checker_name.

        Parameters
        ----------
        name : str
            - The name the desired plugin.
        rebuild_registries: bool (default=None)
            - Whether or not to rebuild the registries if get_plugin fails. If set to
              None, default to what we have set in geoips.filenames.base_paths, which
              defaults to True. If specified, use the input value of rebuild_registries,
              which should be a boolean value. If rebuild registries is true and
              get_plugin fails, rebuild the plugin registry, call then call
              get_plugin once more with rebuild_registries toggled off, so it only gets
              rebuilt once.
        """
        plug = super().get_plugin(name, rebuild_registries)
        if self.valid_plugin(plug):
            return plug

    def valid_plugin(self, plugin):
        """Check the validity of the supplied output_checker plugin."""
        if (
            not hasattr(plugin.module, "outputs_match")
            or not hasattr(plugin.module, "correct_file_format")
            or not hasattr(plugin.module, "call")
        ):
            raise ValidationError(
                "The plugin returned is missing one or more of the following functions."
                "\n[outputs_match, correct_file_format, call]. Please create those "
                "before using this plugin."
            )
        return True


output_checkers = OutputCheckersInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `list interface` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import json
from importlib import resources
from pathlib import Path
import pytest

from geoips import interfaces
from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsListInterface(BaseCliTest):
    """Unit Testing Class for List Interface Command."""

    @property
    def command_combinations(self):
        """A list of every possible call signature for the GeoipsListInterface command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            self._cmd_list = []
            base_args = ["geoips", "list"]
            alias_args = ["geoips", "ls"]
            for argset in [base_args, alias_args]:
                for pkg_name in self.plugin_package_names + ["all"]:
                    for interface_name in interfaces.__all__:
                        interface_name = interface_name.replace("_", "-")
                        if pkg_name != "all":
                            args = argset + [interface_name, "-p", pkg_name]
                        else:
                            args = argset + [interface_name]
                        self._cmd_list.append(args)
            # Add argument list with a non-existent interface
            self._cmd_list.append(base_args + ["non_existent_interface"])
            self._cmd_list.append(alias_args + ["non_existent_interface"])
            # Add argument list that utilizes the --column optional arg
            self._cmd_list.append(
                base_args
                + [
                    "algorithms",
                    "--columns",
                    "package",
                    "interface",
                    "plugin_type",
                    "relpath",
                ]
            )
            self._cmd_list.append(
                alias_args
                + [
                    "algorithms",
                    "--columns",
                    "package",
                    "interface",
                    "plugin_type",
                    "relpath",
                ]
            )
            # Add argument list with an existing interface but non-existent package
            self._cmd_list.append(
                base_args + ["algorithms", "-p", "non_existent_package"]
            )
            self._cmd_list.append(
                alias_args + ["algorithms", "-p", "non_existent_package"]
            )
            # Add argument list with an existing interface but w/ conflicting opt args
            self._cmd_list.append(
                base_args + ["readers", "--long", "--columns", "package", "interface"]
            )
            self._cmd_list.append(
                alias_args + ["readers", "--long", "--columns", "package", "interface"]
            )
        return self._cmd_list

    def check_error(self, args, error):
        """Check that the 'geoips list <interface_name> ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        if "--columns" in args and "--long" in args:
            assert (
                "error: argument --columns/-c: not allowed with argument --long/-l"
                in error.replace("\n", "")
            )
        elif args[2] in interfaces.__all__:
            # interface exists, so check that the package name is incorrect
            assert args[-1] not in self.plugin_package_names
            usg_str = (
                f"error: argument --package_name/-p: invalid "
                f"choice: '{args[-1]}' (choose from"
            )
            assert usg_str in error.replace("\n", "")
        else:
            assert args[2] not in interfaces.__all__
        assert (
            "usage: To use, type `geoips list <interface_name>`" in error
            or "usage: To use, type `geoips list <cmd> <sub-cmd>`" in error
        )

    def check_output(self, args, output):
        """Check that the 'geoips list <interface_name> ...' success output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        interface = getattr(interfaces, args[2].replace("-", "_"))
        interface_type = interface.interface_type
        if "No plugins found under" in output and "-p" in args:
            # No plugins were found under the selected interface, within a
            # certain package; ensure that is correct.
            with open(
                Path(resources.files(args[-1]).joinpath("registered_plugins.json")), "r"
            ) as fo:
                plugin_registry = json.load(fo)

            # assert that the provided interface doesn't exist within that package's
            # plugin registry
            assert interface.name not in plugin_registry[interface_type].keys()
        else:
            # Assert that the correct headers exist in the CLI output
            selected_cols = self.retrieve_selected_columns(args)
            headers = {
                "GeoIPS Package": "package",
                "Interface Name": "interface",
                "Interface Type": "plugin_type",
                "Family": "family",
                "Plugin Name": "plugin_name",
                "Source Names": "source_name",
                "Relative Path": "relpath",
            }
            self.assert_correct_headers_in_output(output, headers, selected_cols)


test_sub_cmd = TestGeoipsListInterface()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips list <interface_name> ...' commands.

    This test covers every valid combination of commands for the
    'geoips list <interface_name>' command. We also test invalid commands, to ensure
    that the proper help documentation is provided for those using the command
    incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with
          (ie. ['geoips', 'list', 'algorithms'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Interpolation methods using pyresample routines."""

import logging

import numpy
from pyresample import kd_tree

LOG = logging.getLogger(__name__)

# interface = None indicates to the GeoIPS interfaces that this is not a valid
# plugin, and this module will not be added to the GeoIPS plugin registry.
# This allows including python modules within the geoips/plugins directory
# that provide helper or utility functions to the geoips plugins, but are
# not full GeoIPS plugins on their own.
interface = None


def get_data_box_definition(source_name, lons, lats):
    """Obtain pyresample geometry definitions.

    For use with pyresample based reprojections

    Parameters
    ----------
    source_name : str
        geoips source_name for data type
    lons : ndarray
        Numpy array of longitudes, 0 to 360
    lats : ndarray
        Numpy array of latitudes, -90 to 90
    """
    from geoips.plugins.modules.interpolators.utils.boxdefinitions import (
        MaskedCornersSwathDefinition,
    )
    from geoips.plugins.modules.interpolators.utils.boxdefinitions import (
        PlanarPolygonDefinition,
    )
    from pyresample.geometry import GridDefinition

    if source_name == "abi" and lons.ndim > 1:
        data_box_definition = GridDefinition(
            lons=numpy.ma.array(lons, subok=False),
            lats=numpy.ma.array(lats, subok=False),
        )
    if source_name == "ahi":
        data_box_definition = PlanarPolygonDefinition(
            lons=numpy.ma.array(lons, subok=False),
            lats=numpy.ma.array(lats, subok=False),
        )
    else:
        data_box_definition = MaskedCornersSwathDefinition(
            lons=numpy.ma.array(lons, subok=False),
            lats=numpy.ma.array(lats, subok=False),
        )

    return data_box_definition


def interp_kd_tree(
    list_of_arrays,
    area_definition,
    data_box_definition,
    radius_of_influence,
    interp_type="nearest",
    sigmas=None,
    neighbours=None,
    nprocs=None,
    fill_value=None,
):
    """Interpolate using pyresample's kd_tree.resample_nearest method.

    Parameters
    ----------
    list_of_array : list of numpy.ndarray
        list of arrays to be interpolated
    area_definition : areadef
        pyresample area_definition object of current region of interest.
    data_box_definition : datadef
        pyresample/geoips data_box_definition specifying region covered
        by source data.
    radius_of_influence : float
        radius of influence for interpolation

    Other Parameters
    ----------------
    interp_type : str, default='nearest'
        One of 'nearest' or 'gauss' - kd_tree resampling methods.
    sigmas : int, default=None
        Used for interp_type 'gauss' - multiplication factor for sigmas option:
            * sigmas = [sigmas]*len(list_of_arrays)
    """
    dstacked_arrays = numpy.ma.dstack(list_of_arrays)

    if interp_type == "nearest":
        kw_args = {}
        kw_args["fill_value"] = None
        kw_args["radius_of_influence"] = radius_of_influence
        if nprocs is not None:
            kw_args["nprocs"] = nprocs
        LOG.info("Using interp_type %s", interp_type)
        dstacked_arrays = kd_tree.resample_nearest(
            data_box_definition,
            dstacked_arrays,
            area_definition,
            **kw_args,
        )
    elif interp_type == "gauss":
        kw_args = {}
        kw_args["sigmas"] = [4000] * len(list_of_arrays)
        kw_args["fill_value"] = None
        kw_args["radius_of_influence"] = radius_of_influence

        if sigmas is not None:
            kw_args["sigmas"] = [sigmas] * len(list_of_arrays)
        if neighbours is not None:
            kw_args["neighbours"] = neighbours
        if nprocs is not None:
            kw_args["nprocs"] = nprocs
        if fill_value is not None:
            kw_args["fill_value"] = fill_value

        LOG.info("Using interp_type %s %s", interp_type, sigmas)
        dstacked_arrays = kd_tree.resample_gauss(
            data_box_definition, dstacked_arrays, area_definition, **kw_args
        )
    else:
        raise TypeError("Unknown interp_type {0}, failing".format(interp_type))

    interpolated_arrays = []
    # We are explicitly expecting 2d arrays back, so if 3d, break it down.
    for arrind in range(len(list_of_arrays)):
        if len(dstacked_arrays.shape) == 3:
            interpolated_arrays += [dstacked_arrays[:, :, arrind]]
        else:
            interpolated_arrays += [dstacked_arrays]

    return interpolated_arrays



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS CLI "tree" command.

Single 'geoips tree' command which will display GeoIPS CLI commands up to a --max-depth
value.
"""

from colorama import Fore, Style

from geoips.commandline.geoips_command import GeoipsExecutableCommand
from geoips.filenames.base_paths import PATHS as gpaths


class GeoipsTree(GeoipsExecutableCommand):
    """Tree Command Class for displaying GeoIPS CLI commands in a tree-like fashion.

    Can be configured to print in a colored fashion, up to a maximum depth, and whether
    or not we want the full command string or just the command name at it's depth.

    Where depth denotes the level of the command you'd like to display.
    I.e. <0> geoips <1> list <2> scripts
    """

    name = "tree"
    command_classes = []

    describe_sectors_outputted = False
    list_sectors_outputted = False

    @property
    def cmd_aliases(self):
        """List of aliases we don't want to include in the tree.

        Without this attribute there will be an entry for each alias of a command name.
        So for 'list' (or 'ls'), there will be two 'geoips list' entries which we don't
        want.
        """
        if not hasattr(self, "_cmd_aliases"):
            self._cmd_aliases = []
            for cmd_name in self.alias_mapping:
                for alias in self.alias_mapping[cmd_name]:
                    # Aliases below are names of actual commands and we need to deal
                    # with this using conditionals in 'print_tree'
                    if alias not in ["sector"]:
                        self._cmd_aliases.append(alias)
        return self._cmd_aliases

    @property
    def top_level_parser(self):
        """The parser associated with the top level 'geoips' command."""
        if not hasattr(self, "_top_level_parser"):
            # Get 'geoips' parser
            self._top_level_parser = self.parent.parser
        return self._top_level_parser

    @property
    def level_to_color(self):
        """A color-mapping dictionary based on the depth of the command tree."""
        if not hasattr(self, "_level_to_color"):
            self._level_to_color = {
                0: Fore.RED,
                1: Fore.YELLOW,
                2: Fore.BLUE,
                3: Fore.MAGENTA,
            }
        return self._level_to_color

    @property
    def cmd_line_url(self):
        """The url to the GeoIPS documentation for the GeoIPS CLI."""
        if not hasattr(self, "_cmd_line_url"):
            self._cmd_line_url = (
                f"{gpaths['GEOIPS_DOCS_URL']}{r'userguide/command_line.html'}"
            )
        return self._cmd_line_url

    def link(self, uri, label=None):
        """Hyperlink the provided uri alongside 'label' if applicable.

        If label is not provided, just return a hyperlink with 'url' as it's text.

        Parameters
        ----------
        uri: str
            - The uniform resource identifier which maps to a certain webpage.
        label: str (default = None)
            - Optional text to hyperlink to.

        Returns
        -------
        hyperlink: str
            - Text that has been hyperlinked to 'uri'
        """
        if label is None:
            label = uri

        parameters = ""
        # OSC 8 ; params ; URI ST <name> OSC 8 ;; ST
        escape_mask = "\033]8;{};{}\033\\{}\033]8;;\033\\"
        hyperlink = escape_mask.format(parameters, uri, label)

        return hyperlink

    def add_arguments(self):
        """Add arguments to the tree-subparser for the Tree Command."""
        self.parser.add_argument(
            "--max-depth",
            default=2,
            type=int,
            help=(
                "The depth of the command tree to print out."
                "Where depth denotes the level of the command you'd like to display."
                "\nI.e. <0> geoips <1> list <2> scripts"
            ),
        )
        self.parser.add_argument(
            "--color",
            default=False,
            action="store_true",
            help="Whether or not we want to print the tree in a colored fashion.",
        )
        self.parser.add_argument(
            "--short-name",
            default=False,
            action="store_true",
            help=(
                "Whether or not we want to print the tree as short names or full names."
                "\nFull names are 'geoips -> geoips config -> geoips config install'."
                "\nShort names are 'geoips -> config -> install'."
            ),
        )

    def print_tree(self, parser, color=False, level=0, max_depth=2, short_name=False):
        """Display GeoIPS CLI commands up to 'max_depth' in a tree-like fashion.

        This function is recursively called up to 'max_depth' to generate that tree of
        commands. If wanted, we can color the output of the tree by depth and we can
        also control whether or not we want the full command string or just the command
        name by depth.

        Parameters
        ----------
        parser: argparse.ArgumentParser
            - The parser associated with the command at the current depth.
        color: bool
            - Whether or not we want the output to be colored by depth.
        level: int
            - The current depth of the command tree (starts at 0).
        max_depth: int
            - How deep we would like to expose the command tree
        short_name:
            - Whether or not we want the full command string or just the command name
              at it's depth.
            - Ex: short_name=False -> 'geoips config install'
            - Ex: short_name=True -> 'install'
        """
        indent = " " * (level * 4)
        split_cmd = parser.prog.split()

        if len(split_cmd) <= 2:
            # split_cmd is either ['geoips'] or ['geoips', '<cmd>']
            cmd_name = " ".join(split_cmd)
        else:
            # split_cmd takes the form of the following (same structure for all cmds):
            # ['To', 'use,', 'type', '`geoips', 'config', '<cmd>', '<sub-cmd>', '...`.', 'install'] # NOQA
            # The case above would result in 'geoips config install'
            cmd_name = " ".join(split_cmd[3:5] + [split_cmd[-1]])[1:]

        if (
            cmd_name == "geoips describe sectors" and self.describe_sectors_outputted
        ) or (cmd_name == "geoips list sectors" and self.list_sectors_outputted):
            # Due to how we've structured aliases, these commands need to be manually
            # checked that they've only ran once. This is caused by 'geoips test sector'
            # command, where 'sector is the actual name of a command so we can't include
            # it in the 'cmd_aliases' property of this class.
            return

        if short_name:
            # Just grab the exact name of the command.
            # Ie. 'geoips' or 'config' or 'install'
            cmd_txt = f"{split_cmd[-1]}"
        else:
            cmd_txt = cmd_name

        # Commenting these portions out until we can get the documentation build and
        # make this easily testable. The introduction of Class Factories for commands
        # has made this much harder to implement.

        # url = f"{self.cmd_line_url}#{cmd_name.replace(' ', '-')}"
        # hyperlink = self.link(url, cmd_txt)
        hyperlink = cmd_txt

        if color:
            # print the command tree in a colored fashion
            color = self.level_to_color.get(level, Fore.WHITE)
            print(f"{indent}{color}{hyperlink}{Style.RESET_ALL}")
        else:
            # Otherwise just print the string in a normal fashion
            print(f"{indent}{hyperlink}")

        if cmd_name == "geoips describe sectors":
            self.describe_sectors_outputted = True
        elif cmd_name == "geoips list sectors":
            self.list_sectors_outputted = True

        if (
            hasattr(parser, "_subparsers")
            and parser._subparsers
            and level + 1 <= max_depth
        ):
            # If the parser has subparsers, subparsers is not None, and we're not
            # exceeding the max depth specified, then call print_tree again.
            subparsers_action = parser._subparsers._group_actions[0]
            for choice, subparser in subparsers_action.choices.items():
                if choice not in self.cmd_aliases:
                    self.print_tree(
                        subparser,
                        color=color,
                        level=level + 1,
                        max_depth=max_depth,
                        short_name=short_name,
                    )

    def __call__(self, args):
        """Run the `geoips tree <opt_args>` command.

        Parameters
        ----------
        args: Namespace()
            - The argument namespace to parse through
        """
        print()
        depth = args.max_depth
        if depth < 0:
            raise self.parser.error(
                f"Invalid depth value: {depth}. Must be greater than or equal to 0."
            )
        self.print_tree(
            self.top_level_parser,
            color=args.color,
            max_depth=depth,
            short_name=args.short_name,
        )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""TC trackfile parser for flat text sectorfiles containing current active storms.

These files contain no storm history, only the currently active storm locations.
Potentially useful for real-time processing.

10S JOSHUA 210120 1200 21.8S 78.1E SHEM 20 1007
12S ELOISE 210120 1800 15.6S 44.9E SHEM 35 1001
92S INVEST 210120 1800 14.9S 120.8E SHEM 30 1002
93S INVEST 210120 1800 12.6S 98.5E SHEM 30 1003
"""
import logging

LOG = logging.getLogger(__name__)

interface = "sector_metadata_generators"
family = "tc"
name = "tc_sector_file_parser"


def call(trackfile_name):
    """TC trackfile parser for flat text sectorfiles containing current active storms.

    These files contain no storm history, only the currently active storm locations.
    Potentially useful for real-time processing.

    Parameters
    ----------
    trackfile_name : str
        Flat text sector file name containing all currently active storm locations,
        formatted as follows:
        * 10S JOSHUA 210120 1200 21.8S 78.1E SHEM 20 1007
        * 12S ELOISE 210120 1800 15.6S 44.9E SHEM 35 1001
        * 92S INVEST 210120 1800 14.9S 120.8E SHEM 30 1002
        * 93S INVEST 210120 1800 12.6S 98.5E SHEM 30 1003

    Returns
    -------
    list
        List of Dictionaries of storm metadata fields from each storm location
        in the flat text sector file

    See Also
    --------
    :ref:`api`
        Valid fields can be found in geoips.sector_utils.utils.SECTOR_INFO_ATTRS
    """
    # These flat text sectorfiles contain all storms -
    # no overall final_storm_name or tc_year
    final_storm_name = None
    tc_year = None

    flatsf_lines = open(trackfile_name).readlines()
    all_fields = []

    for line in flatsf_lines:
        if not line.strip():
            continue
        curr_fields = parse_flat_sectorfile_line(
            line, trackfile_name, parser_name="flat_sectorfile_parser"
        )
        # Was previously RE-SETTING finalstormname here.
        # That is why we were getting incorrect final_storm_name fields
        all_fields += [curr_fields]
    return all_fields, final_storm_name, tc_year


def parse_flat_sectorfile_line(
    line, source_filename, parser_name="flat_sectorfile_parser"
):
    """Retrieve the storm information from the current line from the deck file.

    Parameters
    ----------
    line : str
        Current line from the deck file including all storm information
        * 10S JOSHUA 210120 1200 21.8S 78.1E SHEM 20 1007

    Returns
    -------
    dict
        Dictionary of the fields from the current storm location from the deck file

    See Also
    --------
    :ref:`api`
        Valid fields can be found in geoips.sector_utils.utils.SECTOR_INFO_ATTRS
    """
    from datetime import datetime

    parts = [part.strip() for part in line.split()]
    LOG.info(len(parts))
    if len(parts) < 8:
        raise ValueError(
            "Incorrectly formatted flat text sectorfile - fewer than 8 fields"
        )

    basin_ids = {
        "L": "AL",
        "Q": "SL",
        "A": "IO",
        "B": "IO",
        "S": "SH",
        "P": "SH",
        "C": "CP",
        "E": "EP",
        "W": "WP",
    }

    fields = {}
    # 10S JOSHUA 210120 1200 21.8S 78.1E SHEM 20 1007
    fields["deck_line"] = line.strip()
    fields["storm_basin"] = basin_ids[parts[0][-1]]
    fields["storm_num"] = int(parts[0][:-1])
    fields["synoptic_time"] = datetime.strptime(parts[2] + parts[3], "%y%m%d%H%M")
    fields["aid_type"] = "BEST"
    fields["clat"] = NSEW_to_float(parts[4])
    fields["clon"] = NSEW_to_float(parts[5])
    fields["vmax"] = float(parts[7])
    fields["pressure"] = float(parts[8])

    fields["storm_name"] = parts[1]
    fields["final_storm_name"] = fields["storm_name"]
    fields["storm_year"] = get_storm_year(
        fields["storm_basin"],
        current_month=int(parts[2][2:4]),
        current_year=int("20" + parts[2][0:2]),
    )
    fields["parser"] = parser_name
    fields["source_filename"] = source_filename

    return fields


def NSEW_to_float(lat_lon_val):
    """Convert lat/lon values with NSEW identifiers to positive or negative floats.

    Parameters
    ----------
    lat_lon_val : str
        Latitude or longitude value as a string,
        with hemisphere specified by NSEW identifiers

    Returns
    -------
    float
        Latitude or Longitude value as a float.
    """
    if "S" in lat_lon_val or "W" in lat_lon_val:
        lat_lon_val = -float(lat_lon_val.replace("S", "").replace("W", ""))
    else:
        lat_lon_val = float(lat_lon_val.replace("N", "").replace("E", ""))

    return lat_lon_val


def get_storm_year(storm_basin, current_month, current_year):
    """Ensure correct storm_year is applied.

    For Southern Hemisphere storms that initiate late in the year,
    the storm year identifier is for the following year.

    Parameters
    ----------
    storm_basin : str
        basin of current storm, one of SH, AL, EP, CP, WP, IO
    current_month : int
        Current month of storm location
    current_year : int
        Current year of storm location

    Returns
    -------
    int
        Storm year identifier. current year, unless SH storm later than June,
        then current year + 1
    """
    storm_year = current_year
    if storm_basin == "SH" and current_month > 6:
        storm_year = current_year + 1
    return storm_year



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test all YAML plugins."""
import pytest
import yaml
from importlib import resources, metadata

from geoips.interfaces.base import YamlPluginValidator
from geoips.interfaces.yaml_based.products import ProductsPluginValidator
from geoips.pydantic.workflows import WorkflowPluginModel

validator = YamlPluginValidator()
product_validator = ProductsPluginValidator()


def yield_plugins():
    """Yield plugins."""
    plugin_packages = metadata.entry_points(group="geoips.plugin_packages")
    for pkg in plugin_packages:
        fpath = resources.files(pkg.name) / "plugins/yaml"
        plugin_files = fpath.rglob("*.yaml")
        for pf in plugin_files:
            yield pf


def gen_label(val):
    """Generate the yaml name for pytest ids."""
    return val.name


@pytest.mark.parametrize("plugin", yield_plugins(), ids=gen_label)
def test_is_plugin_valid(plugin):
    """Test if plugin is valid."""
    with open(plugin, "r") as fo:
        docs = list(yaml.safe_load_all(fo))
    for rplugin in docs:
        if rplugin["interface"] == "products":
            product_validator.validate(rplugin)
        elif rplugin["interface"] == "workflows":
            WorkflowPluginModel(**rplugin)
        else:
            validator.validate(rplugin)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module containing user-specified IR-BD algorithm colormap."""

import logging

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "IR_BD"


def call(data_range=[-90.0, 40.0]):
    """Colormap for displaying algorithms/visir/IR_BD.py processed data.

    Parameters
    ----------
    data_range : list of float, default=[-90, 40]
        * Min and max value for colormap.
        * Ensure the data range matches the range of the algorithm specified for
          use with this colormap
        * This colormap MUST include -90 and 40

    Returns
    -------
    mpl_colors_info : dict
        Dictionary of matplotlib plotting parameters, to ensure consistent
        image output
    """
    # for Infrared-Dvorak images at 11 um.  Unit: Celsius
    #     This special black-white color scheme is designed for TC intensity analysis
    min_tb = int(data_range[0])
    max_tb = int(data_range[1])

    if min_tb > -90 or max_tb < 40:
        raise ("Infrared TB range must include -90 and 40")

    from geoips.image_utils.colormap_utils import create_linear_segmented_colormap

    transition_vals = [
        (min_tb, -80.01),
        (-80, -75.01),
        (-75, -69.01),
        (-69, -63.01),
        (-63, -53.01),
        (-53, -41.01),
        (-41, -30.01),
        (-30, 9),
        (9.01, 28),
        (28.01, max_tb),
    ]
    LOG.info("inside util= %s", max_tb)

    # matching TerraScan Color scheme: noaa_bd_151
    transition_colors = [
        ("#555555", "#555555"),
        ("#878787", "#878787"),
        ("white", "white"),
        ("black", "black"),
        ("#A0A0A0", "#A0A0A0"),
        ("#6E6E6E", "#6E6E6E"),
        ("#3C3C3C", "#3C3C3C"),
        ("#C9C9C9", "#6D6D6D"),
        ("#F7F7F7", "#030303"),
        ("black", "black"),
    ]

    # special selection of label

    ticks = [min_tb, -80, -75, -69, -63, -53, -41, -30, -20, -10, 0, 9, 15, 28, max_tb]

    # selection of min and max values for colormap if needed
    min_tb = transition_vals[0][0]
    max_tb = transition_vals[-1][1]
    ticks = ticks + [int(max_tb)]

    LOG.info("Setting cmap")
    mpl_cmap = create_linear_segmented_colormap(
        "IRBD_cmap", min_tb, max_tb, transition_vals, transition_colors
    )

    LOG.info("Setting norm")
    from matplotlib.colors import Normalize

    mpl_norm = Normalize(vmin=min_tb, vmax=max_tb)

    cbar_label = r"11$\mu$m BT ($^\circ$C)"

    # Must be uniform or proportional, None not valid for Python 3
    cbar_spacing = "proportional"
    mpl_tick_labels = None
    mpl_boundaries = None

    # from geoips.image_utils.mpl_utils import create_colorbar
    # only create colorbar for final imagery
    # cbar = create_colorbar(fig, mpl_cmap, mpl_norm, ticks, cbar_label=cbar_label)
    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": mpl_norm,
        "cbar_ticks": ticks,
        "cbar_tick_labels": mpl_tick_labels,
        "cbar_label": cbar_label,
        "boundaries": mpl_boundaries,
        "cbar_spacing": cbar_spacing,
        "colorbar": True,
        "cbar_full_width": True,
    }

    # return cbar, min_tb, max_tb
    return mpl_colors_info



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Takes in a dictionary of xarrays and converts to xarray datatree."""


import logging
from xarray import DataTree

LOG = logging.getLogger(__name__)


def xarray_to_datatree(xarray_dict):
    """Convert a flat (non nested) dictionary of xarrays to DataTree format."""
    tmp_meta = xarray_dict.pop("METADATA")
    xarray_datatree = DataTree(
        name="METADATA", data=tmp_meta, children=DataTree.from_dict(xarray_dict)
    )

    return xarray_datatree



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utilities for working with YAML sector specifications."""

import logging

LOG = logging.getLogger(__name__)


def area_def_to_yamldict(area_def):
    """Convert passed pyresample AreaDefinition to a valid YAML dictionary."""
    yamldict = {}
    sectorname = str(area_def.area_id)
    yamldict[sectorname] = {}
    yamldict = add_sectorinfo_to_yamldict(
        yamldict, sectorname, dict(area_def.sector_info)
    )
    yamldict = add_description_to_yamldict(
        yamldict,
        sectorname,
        sector_type=area_def.sector_type,
        sector_start_datetime=area_def.sector_start_datetime,
        info_dict=dict(area_def.sector_info),
    )

    # The section below was commented out as it is not used by GeoIPS at this time, and
    # the function didn't work since it included errors. 9/27/23

    # yamldict = add_projection_to_yamldict(
    #     yamldict,
    #     sectorname,
    #     area_def.proj_dict["proj"],
    #     area_def.proj_dict["lat_0"],
    #     area_def.proj_dict["lon_0"],
    #     center_x=0,
    #     center_y=0,
    #     pix_x=area_def.x_size,
    #     pix_y=area_def.y_size,
    #     pix_width_m=area_def.pixel_size_x,
    #     pix_height_m=area_def.pixel_size_y,
    # )
    return yamldict


def area_def_to_yamlfile(area_def, out_fname):
    """Write pyresample AreaDefinition out as a valid YAML dictionary."""
    yamldict = area_def_to_yamldict(area_def)
    return write_yamldict(yamldict, out_fname=out_fname)


def write_yamldict(yamldict, out_fname, force=False, replace_geoips_paths=False):
    """Write yamldict to out_fname.

    Parameters
    ----------
    yamldict : dict
        Dictionary to write out to YAML file
    out_fname : str
        Output filename to write YAML dict to
    force : bool, default=False
        If True, overwrite existing file.
    replace_geoips_paths: bool, default=False
        If True, replace full path with appropriate environment variable in YAML output

    Returns
    -------
    str
        Path to output file if successfully produced
    """
    from geoips.filenames.base_paths import make_dirs
    from geoips.geoips_utils import replace_geoips_paths_in_dict
    from os.path import dirname, exists
    import yaml

    if replace_geoips_paths:
        dump_yamldict = replace_geoips_paths_in_dict(yamldict)
    # If we aren't replacing the geoips paths, just use the original yaml dict.
    else:
        dump_yamldict = yamldict

    make_dirs(dirname(out_fname))
    if not exists(out_fname) or force:
        with open(out_fname, "w") as fobj:
            LOG.info("SUCCESS Writing out yaml file %s", out_fname)
            yaml.safe_dump(dump_yamldict, fobj, default_flow_style=False)
            return [out_fname]
    else:
        LOG.info(
            "SKIPPING %s already exists, delete it if you need to recreate", out_fname
        )
        return []


def add_dynamic_datetime_to_yamldict(
    yaml_dict, sectorname, sector_start_datetime, sector_end_datetime
):
    """Add passed dynamic datetime info to passed YAML dictionary."""
    yaml_dict[sectorname]["sector_start_datetime"] = sector_start_datetime
    yaml_dict[sectorname]["sector_end_datetime"] = sector_end_datetime
    return yaml_dict


def add_description_to_yamldict(
    yaml_dict, sectorname, sector_type, sector_start_datetime=None, info_dict=None
):
    """Add passed sector description information to passed YAML dictionary."""
    yaml_dict[sectorname]["sector_type"] = sector_type
    if sector_type == "static":
        yaml_dict[sectorname]["description"] = sectorname
    if sector_type == "tc":
        yaml_dict[sectorname]["sector_type"] = sector_type
        yaml_dict[sectorname]["description"] = "TC{0} {1}{2} {3} {4}".format(
            info_dict["storm_year"],
            info_dict["storm_basin"],
            info_dict["storm_num"],
            info_dict["storm_name"],
            str(info_dict["synoptic_time"]),
        )
    if sector_type in ["pyrocb", "atmosriver", "volcano"]:
        sector_start_datetime_str = yaml_dict[sectorname][
            "sector_start_datetime"
        ].strftime("%Y%m%dT%HZ")
        yaml_dict[sectorname]["sector_type"] = sector_type
        yaml_dict[sectorname]["description"] = "{0} at {1}".format(
            sectorname, sector_start_datetime_str
        )
    return yaml_dict


def add_sectorinfo_to_yamldict(yaml_dict, sectorname, sector_info_dict):
    """Add sector_info dictionary to YAML dictionary."""
    yaml_dict[sectorname]["sector_info"] = sector_info_dict
    return yaml_dict


# The function below were commented out as they included errors, and were not used
# by GeoIPS at this time. 9/27/23

# def add_projection_to_yamldict(
#     yaml_dict,
#     sectorname,
#     center_lat,
#     center_lon,
#     center_x=0,
#     center_y=0,
#     template_yaml=None,
# ):
#     """Add projection information to YAML dictionary."""
#     LOG.info("add_projection_to_yamldict - update to template_yaml")

#     yaml_dict[sectorname]["projection"] = {}
#     yaml_dict[sectorname]["projection"]["proj"] = proj
#     yaml_dict[sectorname]["projection"]["a"] = 6371228.0
#     yaml_dict[sectorname]["projection"]["units"] = "m"
#     yaml_dict[sectorname]["projection"]["lat_0"] = center_lat
#     yaml_dict[sectorname]["projection"]["lon_0"] = center_lon
#     yaml_dict[sectorname]["center"] = [center_x, center_y]
#     yaml_dict[sectorname]["resolution"] = [pix_width_m, pix_height_m]
#     # yaml_dict[sectorname]['shape'] = [pix_x, pix_y]
#     yaml_dict[sectorname]["shape"] = {}
#     yaml_dict[sectorname]["shape"]["width"] = pix_x
#     yaml_dict[sectorname]["shape"]["height"] = pix_y
#     # This only works because it is square!!
#     yaml_dict[sectorname]["area_extent"] = {
#         "lower_left_xy": [
#             center_x - (pix_x * pix_width_m / 2),
#             center_y - (pix_y * pix_height_m / 2),
#         ],
#         "upper_right_xy": [
#             center_x + (pix_x * pix_width_m / 2),
#             center_y + (pix_y * pix_height_m / 2),
#         ],
#     }
#     return yaml_dict



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Routines for converting between units."""

# Python Standard Libraries
import logging

# Installed Libraries

KtoC_conversion = -273.15

LOG = logging.getLogger(__name__)


def unit_conversion(data_array, input_units=None, output_units=None):
    """Convert array in units 'input_units' to units 'output_units'.

    Parameters
    ----------
    data_array : ndarray
        numpy.ndarray or numpy.MaskedArray of data values to be converted
    input_units : str, optional
        Units of input data array, defaults to None
    output_units : str, optional
        Units of output data array, defaults to None

    Returns
    -------
    MaskedArray
        Return numpy.ma.MaskedArray, with units converted
        from 'input_units' to 'output_units'
    """
    if input_units and output_units and input_units != output_units:
        from geoips.data_manipulations.corrections import apply_offset

        valid_units = ["Kelvin", "celsius", "kts", "m s-1"]
        if input_units not in valid_units:
            raise ValueError(f"Input units must be one of {valid_units}")
        if output_units not in valid_units:
            raise ValueError(f"Output units must be one of {valid_units}")

        if input_units == "Kelvin" and output_units == "celsius":
            data_array = apply_offset(data_array, KtoC_conversion)

        if input_units == "celsius" and output_units == "Kelvin":
            data_array = apply_offset(data_array, -KtoC_conversion)

        if input_units == "m s-1" and output_units == "kts":
            data_array *= 1.94384

        if input_units == "kts" and output_units == "m s-1":
            data_array *= 0.514444

    return data_array



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Simple script to list available plugins for each interface."""
import warnings
from geoips.interfaces.base import BaseInterface
from geoips import interfaces
from geoips.commandline.log_setup import setup_logging
from geoips.commandline.args import get_argparser, check_command_line_args

# Always actually raise DeprecationWarnings
# Note this SO answer https://stackoverflow.com/a/20960427
warnings.simplefilter("always", DeprecationWarning)


def main():
    """Script to list all modules available in the current geoips instantiation."""
    supported_args = ["logging_level"]

    # Use standard GeoIPS argparser - use standard formatting for standard args.
    parser = get_argparser(
        arglist=supported_args,
        description="List available plugins within the current geoips instantiation.",
    )
    # Add our specific "interfaces" argument.
    parser.add_argument(
        "interfaces",
        nargs="*",
        default=None,
        help="""List of interfaces to include.  If None, include all interfaces.""",
    )

    # Get the dictionary of command line args.
    COMMAND_LINE_ARGS = parser.parse_args().__dict__
    # Check included arguments for appropriate formatting / type.
    check_command_line_args(supported_args, COMMAND_LINE_ARGS)

    # Setup logging at the requested logging level.
    LOG = setup_logging(logging_level=COMMAND_LINE_ARGS["logging_level"])

    # If "interfaces" is not specified command line, then include plugins for
    # all interfaces.
    if not COMMAND_LINE_ARGS["interfaces"]:
        curr_interfaces = interfaces.__dict__.values()
    else:
        curr_interfaces = [
            interfaces.__dict__[curr_interface]
            for curr_interface in COMMAND_LINE_ARGS["interfaces"]
        ]

    # Print just the interface name and plugin name for each plugin in each
    # requested interface.
    for curr_interface in curr_interfaces:
        if (type(curr_interface) is BaseInterface) or not isinstance(
            curr_interface, BaseInterface
        ):
            continue

        # This prints a "horizontal line" character U+2015, not a dash
        LOG.interactive("―" * len(curr_interface.name))
        LOG.interactive(f"{curr_interface.name}")
        LOG.interactive("―" * len(curr_interface.name))
        plugin_names = sorted([plg.name for plg in curr_interface.get_plugins()])
        LOG.interactive(", ".join(plugin_names))


if __name__ == "__main__":
    main()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Gridline Annotator interface module."""

from geoips.interfaces.base import BaseYamlInterface


class GridlineAnnotatorsInterface(BaseYamlInterface):
    """Interface for gridline annotator plugins."""

    name = "gridline_annotators"

    # Keyword arguments that can be used to adjust the style of lat/lon labels applied
    # to annotated imagery. See geoips.image_utils.maps:draw_gridlines for how this is
    # implemented
    label_kwargs = {
        "alpha": ["number"],
        "backgroundcolor": ["string"],
        "color": ["string"],
        "fontfamily": ["string"],
        "fontsize": ["number", "string"],
        "fontstretch": ["integer", "string"],
        "fontstyle": ["string"],
        "fontvariant": ["string"],
        "fontweight": ["integer", "string"],
        "linespacing": ["number"],
        "mouseover": ["boolean"],
        "position": ["array"],
        "rotation": ["number", "string"],
        "rotation_mode": ["string"],
        "snap": ["boolean"],
        "wrap": ["boolean"],
        "zorder": ["number"],
    }


gridline_annotators = GridlineAnnotatorsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Generate standard pyresample area definitions given center coordinates.

Given desired center lat/lon, projection, resolution, and shape, return a
valid pyresample area definition object.
"""

interface = "sector_spec_generators"
family = "area_definition"
name = "center_coordinates"


def set_clat_clon_proj_info(
    clat, clon, pr_proj, num_samples, num_lines, pixel_width, pixel_height
):
    """Create standard proj4 dictionary from passed projection information."""
    width_m = num_samples * pixel_width
    height_m = num_lines * pixel_height

    proj4_dict = {
        "proj": pr_proj,
        "a": 6371228.0,
        "lat_0": clat,
        "lon_0": clon,
        "units": "m",
    }
    area_left = -width_m / 2.0
    area_right = width_m / 2.0
    area_bot = -height_m / 2.0
    area_top = height_m / 2.0
    area_extent = (area_left, area_bot, area_right, area_top)
    return proj4_dict, area_extent


def call(
    area_id,
    long_description,
    clat,
    clon,
    projection,
    pixel_width,
    pixel_height,
    num_samples,
    num_lines,
):
    """Create area definition using clat, clon, resolution, and shape."""
    proj4_dict, area_extent = set_clat_clon_proj_info(
        clat=clat,
        clon=clon,
        pr_proj=projection,
        num_samples=num_samples,
        num_lines=num_lines,
        pixel_width=pixel_width,
        pixel_height=pixel_height,
    )

    # Create the AreaDefinition object from given fields.
    # We are currently relying on Sector objects
    # for some information - need to decide how to handle this directly.
    # This is area_id= then name= for Python2, area_id= then description= for Python3
    from pyresample import AreaDefinition

    try:
        # Backwards compatibility for Python 2 version of pyresample
        area_def = AreaDefinition(
            area_id,
            long_description,
            proj_id="{0}_{1}".format(proj4_dict["proj"], area_id),
            proj_dict=proj4_dict,
            x_size=num_samples,
            y_size=num_lines,
            area_extent=area_extent,
        )
    except TypeError:
        area_def = AreaDefinition(
            area_id,
            long_description,
            proj_id="{0}_{1}".format(proj4_dict["proj"], area_id),
            projection=proj4_dict,
            width=num_samples,
            height=num_lines,
            area_extent=area_extent,
        )
    return area_def



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Matplotlib-based unprojected image output."""

import logging
from os.path import basename, dirname, join

import matplotlib.pyplot as plt
import matplotlib

from geoips.image_utils.mpl_utils import save_image

matplotlib.use("agg")

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "unprojected"
name = "unprojected_image"


def call(
    xarray_obj,
    product_name,
    output_fnames,
    product_name_title=None,
    mpl_colors_info=None,
    x_size=None,
    y_size=None,
    savefig_kwargs=None,
    is_3d=False,
):
    """Plot unprojected image to matplotlib figure."""
    if savefig_kwargs is None:
        # Default to no arguments, empty dictionary.  Will result in masked background
        savefig_kwargs = {}
        # black background
        # savefig_kwargs = {'facecolor': 'k'}

    if product_name_title is None:
        product_name_title = product_name
    if x_size is None and y_size is None:
        x_size = xarray_obj[product_name].shape[1]
        y_size = xarray_obj[product_name].shape[0]
    elif y_size is None:
        ratio = float(x_size) / float(xarray_obj[product_name].shape[1])
        y_size = float(ratio) * float(xarray_obj[product_name].shape[0])
    elif x_size is None:
        ratio = float(y_size) / float(xarray_obj[product_name].shape[0])
        x_size = float(ratio) * float(xarray_obj[product_name].shape[1])

    rc_params = matplotlib.rcParams
    dpi = rc_params["figure.dpi"]

    image_width = float(x_size) / dpi
    image_height = float(y_size) / dpi

    fig = plt.figure(facecolor="none")
    fig.set_size_inches(image_width, image_height)
    main_ax = plt.Axes(fig, [0, 0, 1, 1])
    main_ax.set_axis_off()
    fig.add_axes(main_ax)

    # This needs to be done as output_formatter_kwargs can only be specified as strings
    # (to my knowledge at least)
    is_3d = str(is_3d) == "True"

    if is_3d:
        slices = [
            xarray_obj[product_name].data[slice]
            for slice in range(xarray_obj[product_name].data.shape[0])
        ]
    else:
        slices = [xarray_obj[product_name].data]
    for slice_idx, slice_data in enumerate(slices):
        main_ax.clear()
        main_ax.imshow(
            slice_data,
            norm=mpl_colors_info["norm"],
            cmap=mpl_colors_info["cmap"],
        )

        success_outputs = []
        for fname in output_fnames:
            if is_3d:
                # This is generic for overcast data, on the order of (level) * 0.5 km.
                # I.e. if level == 5, lvl_str == '02_50.'. This is an easy way to denote
                # what height each image corresponds to, though I'm not sure if we
                # should hardcode this string here. Maybe create a new filename
                # formatter for that data or we can add a new argument to this plugin
                # which handles fname conventions for 3D data. Such as a function which
                # produces a pre-prended string to the corresponding fname.
                lvl_km = str(slice_idx * 0.5).split(".")
                lvl_str = f"{lvl_km[0].zfill(2)}_{lvl_km[1]}0"
                suffix = f".{basename(fname).split('.')[-1]}"
                # This is expected from ovcst_fname filename_formatter
                if "ovcst" in basename(fname):
                    # OVCST_<product_name>_<datetime>.png
                    # <datetime> fmt = yyyymmddhhnnss
                    datetime = basename(fname).split("_")[-1][: -len(suffix)]
                    date = datetime[:8]
                    time = datetime[8:]
                # This is expected from basic_fname filename_formatter
                else:
                    date, time = basename(fname).split(".")[0:2]
                # Making a directory with date.time/fname as there are 40 images per
                # file for OVERCAST data
                final_fname = join(
                    dirname(fname),
                    f"{date}.{time}",
                    f"{basename(fname)[:-len(suffix)]}_{lvl_str}{suffix}",
                )
            else:
                final_fname = fname
            LOG.info("Plotting %s with plt", fname)
            # This just handles cleaning up the axes, creating directories, etc
            success_outputs += save_image(
                fig,
                final_fname,
                is_final=False,
                image_datetime=xarray_obj.start_datetime,
                savefig_kwargs=savefig_kwargs,
            )

    return success_outputs



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Windsat binary data reader.

This code is designed to read windsat sdr binary data (idr37) file for windsat
37 GHz products in GEOIPS environments. the input file name is something alike
US058SORB-BINspp.wndmi_fws_d20191126_s134102_e153244_r87467_cfnmoc.idr37.

V1.0: initial version.  Song Yang, NRL-MRY, 01/08/2020

errflag is the important parameter of the windsat edr dataset.  It is a 32-bit
integer which describes what is the current data point status.  Here are the
meaning of each bit::

    0-7: Wilheit rain flag
      8: forward/aft scan (bit set to 1 for forward part of scan, 0 for aft scan )
      9: ascending/descending pass flag (1 for ascending, 0 for descending)
     10: Warm load flag
     11: Warm load gains applied (1 = gains applied, 0 = gains not applied)
     12: Glare angle invalid because no 1 vector or LOS doesn't pierce earth
     13-18: Glare angle (0 to 30 represents angles of 0 to 60 degree in increments
            of 2 deg; 31 represents angles .gt. 60 deg; 32 represents invalid
            glare angle)
     19: Cold load flag. If set to 1 the VH channel data had to be corrected due
         to interference in the cold load signal, such as the moon or a
         geostationary satellite.
     20: Gain Saturation flag. Set to 1 when strong RFI causes the gain to change.
         This is set if any TDR saturation flag is set at this frequency
     23: used to hold the rfi flag so that it may be passed on to other structures
         such as the resampling and intermediate structures. Finally RFI is placed
         in the sdr structure.
     other bits are spare

Here is the original sdr record in Fortran::

    type IDRRecord_short
       real(double) :: JD2000           8 bytes
       real, dimension(4)::  stokes     16 bytes
       real :: latitude                 4 bytes
       real :: longitude                4 bytes
       real :: EIA                      4 bytes: earth incidence angle, the angle
                                                 on the ground between vertical and
                                                 the satellite look vector
       real :: PRA                      4 bytes: rotation of the polarization plane
                                                 from true
       real :: CAA                      4 bytes: compass azimuth angle on the ground
       real :: tI45, tIcp, pra45        12 bytes
       integer :: errflag               4 bytes (32-bits integer):a set of bit
                                                flags for data quality and
                                                conditions (above explanation)
       integer :: Scan                  4 bytes: scan line number in the orbit.
                                                 WindSat scans every 1.9 seconds
       integer(int16) :: dcnum          2 bytes: pixel number along the scan,
                                                 called 'downcount number',
                                                 because the highest pixel number
                                                 is measured first.
       integer(int16) :: SurfaceType    2 bytes: legacy SSMI surface type
       integer(int16) :: scanAngle      2 bytes: angle on the ground between the
                                                 flight direction and the look
                                                 direction
       integer(int8) :: water2land ! copied from IDRL record   1 byte:
       integer(int8) :: land2water ! copied from IDRL record   1 byte:
    end type IDRRecord_short

* Gain saturation is when a sudden, large signal causes the gain to change
  quickly and make averaged gain unreliable.
* Forward/Aft is for sensor view position
* The warm load flag indicates that calibration may be unreliable due to solar
  intrusion into the warm load.
* Cold load flags do not mean calibration is unreliable. It's a way for us to
  check the cold load correction algorithm.
* Sun glare is not something to worry about.
* The RFI flag is never set at 37 GHz.
* tI45, tIcp, pra45, scanAngleI, water2land, and land2water aren't commanly used.

   * The first four are for recreating the 6-element pre-Stokes polarization
     vector, and the last two measure coastal contamination

The actual idr37 data record (idr_record) in C::

    typedef struct {
       double jd2000;
       float stokes[4];
       float plat;     lat of earth observation
       float plon;     lon of earth observation
       float eia;      radiance= ~53deg
       float pra;
       float caa;
       float slat;  latitude of satellite position?  not
       float slon;  longitude of satellite position? not
       float salt;  altitude of satellite (meter? km?) not
       int errflag;
       int scan;
       short dcnum;
       short surf;
       float spare;
    idr_record;

    Its total length of idr_record is 72 bytes
"""
# Python Standard Libraries
from datetime import datetime
import logging
import os

# Third-Party Libraries
import numpy as np

# import pandas as pd
import xarray


LOG = logging.getLogger(__name__)

dataset_info = {
    "WINDSAT_SDR_FWD": {
        "ftime_jd2000": "ftime_jd2000",
        "ftb37v": "ftb37v",
        "ftb37h": "ftb37h",
        "flat": "flat",
        "flon": "flon",
        "fsurfaceType": "fsurfaceType",
        "frainFlag": "frainFlag",
        "fasc_des_pass": "fasc_des_pass",
    },
    "WINDSAT_SDR_AFT": {
        "atime_jd2000": "atime_jd2000",
        "atb37v": "atb37v",
        "atb37h": "atb37h",
        "alat": "alat",
        "alon": "alon",
        "asurfaceType": "asurfaceType",
        "arainFlag": "arainFlag",
        "aasc_des_pass": "aasc_des_pass",
    },
}
gvar_info = {
    "WINDSAT_SDR_FWD": {"Latitude": "latitude", "Longitude": "longitude"},
    "WINDSAT_SDR_AFT": {"Latitude": "latitude", "Longitude": "longitude"},
}

interface = "readers"
family = "standard"
name = "windsat_idr37_binary"
source_names = ["windsat"]

# NOTE: Anytime you see a # NOQA comment, this is for flake8 formatting. Unused
# variables are needed in this for moving through the binary file correctly. There is
# fmt: off and fmt: on comments, which prevent black from moving the # NOQA comments.


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read Windsat binary data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # get data time info from input fname
    fname = fnames[0]
    time_s_date = os.path.basename(fname).split("_")[2][1:9]
    time_s_hhmm = os.path.basename(fname).split("_")[3][1:5]
    time_e_hhmm = os.path.basename(fname).split("_")[4][1:5]

    # findout whether this data file crossing boundary of day, month, or year
    time_s_year = time_s_date[0:4]
    time_s_month = time_s_date[4:6]
    time_s_day = time_s_date[6:8]

    # initilization of ending date for this file
    time_e_year = int(time_s_year)
    time_e_month = int(time_s_month)
    time_e_day = int(time_s_day)

    # if time_e_hhmm is less than time_s_hhmm, the data crossed the day
    # bounday and entered the next day

    if (int(time_s_year) % 400 == 0) or (
        (int(time_s_year) % 4 == 0) and (int(time_s_year) % 100 != 0)
    ):
        year_leap = "true"
        LOG.info("%d is a Leap Year" % int(time_s_year))
    else:
        year_leap = "false"
        LOG.info("%d is Not the Leap Year" % int(time_s_year))

    if int(time_e_hhmm) < int(time_s_hhmm):
        time_e_day = int(time_s_day) + 1
        if (
            (int(time_s_month) == 4)
            or (int(time_s_month) == 6)
            or (int(time_s_month) == 9)
        ):  # in Apr, Jun and Sep
            if time_e_day > 30:
                time_e_day == 1  # first day of the next month
                time_e_month == int(time_s_month) + 1
        else:
            if int(time_s_month) == 2:  # in Feb
                if year_leap:  # In a leap year
                    if time_e_day > 29:
                        time_e_day == 1  # first day of the next month
                        time_e_month == int(time_s_month) + 1
                else:  # No leap year
                    if time_e_day > 28:
                        time_e_day == 1  # first day of the next month
                        time_e_month == int(time_s_month) + 1
            else:  # in Jan, Mar, Jul, Aug, Oct, Dec
                if time_e_day > 31:
                    time_e_day == 1  # first day of the next month
                    time_e_month == int(time_s_month) + 1
        if time_e_month > 12:  # in Dec, deterimine whether it crosses the year boundary
            time_e_year == int(time_s_year) + 1

    # convret end_time of the data into strings
    time_e_date = (
        str(time_e_year) + str("%02d" % time_e_month) + str("%02d" % time_e_day)
    )

    # Need to set up time to be read in by the metadata (year and jday are arrays)
    time_start = time_s_date + " " + time_s_hhmm
    time_end = time_e_date + " " + time_e_hhmm

    xarray_obj = xarray.Dataset()

    # Enter metadata
    xarray_obj.attrs["start_datetime"] = datetime.strptime(time_start, "%Y%m%d %H%M")
    xarray_obj.attrs["end_datetime"] = datetime.strptime(time_end, "%Y%m%d %H%M")
    xarray_obj.attrs["source_file_datetimes"] = [xarray_obj.attrs["start_datetime"]]

    xarray_obj.attrs["platform_name"] = "coriolis"
    xarray_obj.attrs["source_name"] = "windsat"
    xarray_obj.attrs["interpolation_radius_of_influence"] = 10000
    xarray_obj.attrs["sample_distance_km"] = 2
    xarray_obj.attrs["data_provider"] = "NRL-NOAA"

    # Passing chans == [] indicates we do not want ANY data, only metadata, so
    # return once metadata is set.
    if metadata_only:
        return {"METADATA": xarray_obj}

    # find out size of the input file (bytes)
    filesize_info = os.stat(fname).st_size
    len_OneRec = 72  # size of record (bytes)?
    rec_tot = filesize_info // len_OneRec  # how many records in the file

    # check for exact data records by mode of the file_size vs data_record
    try:
        good_datafile = filesize_info % len_OneRec
        if good_datafile == 0:
            LOG.info("This is a good windsat idr37 data file")
        else:
            LOG.info("This is not a good windsat idr37 data file:  skipping ....")
            return
    except Exception as resp:
        LOG.info(
            "\tBLANKET EXCEPTION %s: %s >> %s : %s",
            type(resp).__name__,
            str(resp.__doc__),
            str(resp.args),
            "windsat idr37 data does not have even number of data records !! "
            "Skipping...",
        )
        return {"METADATA": xarray_obj}

    bad_value = -999
    # declare data arrays
    try:
        windsat_read = np.ma.zeros(rec_tot)  # initialization of zeros
        np.ma.masked_all_like(windsat_read)
    except Exception as resp:
        LOG.info(
            "\tBLANKET EXCEPTION %s: %s >> %s : %s",
            type(resp).__name__,
            str(resp.__doc__),
            str(resp.args),
            "windsat idr37 data does not have even number of data records !! "
            "Skipping...",
        )
        return {"METADATA": xarray_obj}

    ftime_jd2000 = np.ma.masked_values(windsat_read, bad_value)
    ftb37v = np.ma.masked_values(windsat_read, bad_value)
    ftb37h = np.ma.masked_values(windsat_read, bad_value)
    flat = np.ma.masked_values(windsat_read, bad_value)
    flon = np.ma.masked_values(windsat_read, bad_value)
    fsurfaceType = np.ma.masked_values(windsat_read, bad_value)
    frainFlag = np.ma.masked_values(windsat_read, bad_value)
    fasc_des_pass = np.ma.masked_values(windsat_read, bad_value)

    atime_jd2000 = np.ma.masked_values(windsat_read, bad_value)
    atb37v = np.ma.masked_values(windsat_read, bad_value)
    atb37h = np.ma.masked_values(windsat_read, bad_value)
    alat = np.ma.masked_values(windsat_read, bad_value)
    alon = np.ma.masked_values(windsat_read, bad_value)
    asurfaceType = np.ma.masked_values(windsat_read, bad_value)
    arainFlag = np.ma.masked_values(windsat_read, bad_value)
    aasc_des_pass = np.ma.masked_values(windsat_read, bad_value)

    f1 = open(fname, "rb")

    k = 0
    k2 = 0
    # read in the windsat edr products for all data  points
    for ii in range(rec_tot):  # loop records of this file
        if ii % 10000 == 0:
            LOG.info("Running record number %s of %s", ii, rec_tot)
        try:
            # read in variables using their size (bytes)
            jd2000 = np.frombuffer(f1.read(8), dtype=np.dtype("float64")).byteswap()[
                0
            ]  # sec since 1200Z,01/01/2000
            # get time info for each data point
            # fmt: off
            # timeinfo = pd.datetime(2000, 1, 1, 12) + pd.Timedelta(jd2000, unit="s")
            # time_date = int(
            #     str(timeinfo.year) + str(timeinfo.month) + str(timeinfo.day)
            # )
            # time_hhmm = int(str(timeinfo.hour) + str(timeinfo.minute))
            tb37v, tb37h, tb37info1, tb37info2 = np.frombuffer(
                f1.read(16), dtype=np.dtype("float32")
            ).byteswap()  # K (37v,37h,info1,info2)
            lat = np.frombuffer(f1.read(4), dtype=np.dtype("float32")).byteswap()[
                0
            ]  # deg
            lon = np.frombuffer(f1.read(4), dtype=np.dtype("float32")).byteswap()[
                0
            ]  # deg
            # if lon > 180.0:          # windbarbs needs lon in (-180,180)
            #   lon=lon-360

            eia = np.frombuffer(  # NOQA
                f1.read(4), dtype=np.dtype("float32")
            ).byteswap()[
                0
            ]  # unit in radiance =~53 deg
            pra = np.frombuffer(f1.read(4), dtype=np.dtype("float32")).byteswap()[  # NOQA
                0
            ]
            caa = np.frombuffer(f1.read(4), dtype=np.dtype("float32")).byteswap()[  # NOQA
                0
            ]
            slat = np.frombuffer(  # NOQA
                f1.read(4), dtype=np.dtype("float32")
            ).byteswap()[
                0
            ]  # deg
            slon = np.frombuffer(  # NOQA
                f1.read(4), dtype=np.dtype("float32")
            ).byteswap()[
                0
            ]  # deg
            salt = np.frombuffer(  # NOQA
                f1.read(4), dtype=np.dtype("float32")
            ).byteswap()[
                0
            ]  # meter
            errflag = np.frombuffer(f1.read(4), dtype=np.dtype("int32")).byteswap()[
                0
            ]  # deg
            scanNum = np.frombuffer(f1.read(4), dtype=np.dtype("int32")).byteswap()[  # NOQA
                0
            ]
            downcountNum = np.frombuffer(  # NOQA
                f1.read(2), dtype=np.dtype("int16")
            ).byteswap()[0]
            surfaceType = np.frombuffer(f1.read(2), dtype=np.dtype("int16")).byteswap()[
                0
            ]
            spare = np.frombuffer(  # NOQA
                f1.read(4), dtype=np.dtype("float32")
            ).byteswap()[
                0
            ]  # spare var for space holder

            # fmt: on
            # decode errflag to assign value to approperated variables, i.e.,
            # forward/aft mode, ascending/descending etc
            fore_aft_scan = (
                errflag >> 8 & 1
            )  # =1, foreward scan; =0, aft scan (foreward scan used for image)
            asc_des_pass = errflag >> 9 & 1  # =1, ascending; =0, descending

            # rainflag
            rrflag_tmp = 0
            for j in range(8):
                bit = errflag >> j & 1  # =1, set, =0, not set
                rrflag_tmp += bit * pow(2, j)
            rrflag = rrflag_tmp

            if (
                fore_aft_scan == 1
            ):  # forward scan points  --> will be used for image products
                ftb37v[k] = tb37v
                ftb37h[k] = tb37h
                flat[k] = lat
                flon[k] = lon
                fsurfaceType[k] = surfaceType
                frainFlag[k] = rrflag
                ftime_jd2000[k] = jd2000
                fasc_des_pass[k] = asc_des_pass
                k += 1
            else:  # aft scan points
                atb37v[k2] = tb37v
                atb37h[k2] = tb37h
                alat[k2] = lat
                alon[k2] = lon
                asurfaceType[k2] = surfaceType
                arainFlag[k2] = rrflag
                atime_jd2000[k2] = jd2000
                aasc_des_pass[k2] = asc_des_pass
                k2 += 1

        except Exception as resp:
            LOG.info(
                "\tBLANKET EXCEPTION %s: %s >> %s : %s",
                type(resp).__name__,
                str(resp.__doc__),
                str(resp.args),
                "Failed setting windsat sdr data arrays!! Skipping...",
            )
    f1.close()

    flat = np.ma.masked_values(flat, bad_value)
    flon = np.ma.masked_values(flon, bad_value)
    ftime_jd2000 = np.ma.masked_values(ftime_jd2000, bad_value)
    ftb37v = np.ma.masked_values(ftb37v, bad_value)
    ftb37h = np.ma.masked_values(ftb37h, bad_value)
    fsurfaceType = np.ma.masked_values(fsurfaceType, bad_value)
    frainFlag = np.ma.masked_values(frainFlag, bad_value)
    fasc_des_pass = np.ma.masked_values(fasc_des_pass, bad_value)

    alat = np.ma.masked_values(alat, bad_value)
    alon = np.ma.masked_values(alon, bad_value)
    atime_jd2000 = np.ma.masked_values(atime_jd2000, bad_value)
    atb37v = np.ma.masked_values(atb37v, bad_value)
    atb37h = np.ma.masked_values(atb37h, bad_value)
    asurfaceType = np.ma.masked_values(asurfaceType, bad_value)
    arainFlag = np.ma.masked_values(arainFlag, bad_value)
    aasc_des_pass = np.ma.masked_values(aasc_des_pass, bad_value)

    xarray_sdr_fwd = xarray.Dataset()
    xarray_sdr_aft = xarray.Dataset()
    xarray_sdr_aft = xarray.Dataset()
    xarray_sdr_fwd.attrs = xarray_obj.attrs.copy()
    xarray_sdr_aft.attrs = xarray_obj.attrs.copy()

    timediff = np.datetime64("2000-01-01T12:00:00") - np.datetime64(
        "1970-01-01T00:00:00"
    )
    timestamps = ftime_jd2000.astype("np.datetime64[s]") + timediff
    xarray_sdr_aft["time"] = xarray.DataArray(timestamps)
    xarray_sdr_aft["latitude"] = xarray.DataArray(alat)
    xarray_sdr_aft["longitude"] = xarray.DataArray(alon)
    xarray_sdr_aft["atb37v"] = xarray.DataArray(atb37v)
    xarray_sdr_aft["asurfaceType"] = xarray.DataArray(asurfaceType)
    xarray_sdr_aft["arainFlag"] = xarray.DataArray(arainFlag)
    xarray_sdr_aft["aasc_des_pass"] = xarray.DataArray(aasc_des_pass)

    xarray_sdr_fwd["time"] = xarray.DataArray(timestamps)
    xarray_sdr_fwd["latitude"] = xarray.DataArray(flat)
    xarray_sdr_fwd["longitude"] = xarray.DataArray(flon)
    xarray_sdr_fwd["ftb37v"] = xarray.DataArray(ftb37v)
    xarray_sdr_fwd["fsurfaceType"] = xarray.DataArray(fsurfaceType)
    xarray_sdr_fwd["frainFlag"] = xarray.DataArray(frainFlag)
    xarray_sdr_fwd["fasc_des_pass"] = xarray.DataArray(fasc_des_pass)
    return {"METADATA": xarray_obj, "AFT": xarray_sdr_aft, "FWD": xarray_sdr_fwd}
